package de.fh_zwickau.bumb_l_b.qos_tool;

import java.security.InvalidParameterException;

import org.jbpt.pm.ControlFlow;
import org.jbpt.pm.FlowNode;
import org.jbpt.pm.NonFlowNode;
import org.jbpt.pm.epc.IEpc;

import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper.AttributeType;
import de.fh_zwickau.bumb_l_b.qos_tool.loader.EpcLoaderException;
import de.fh_zwickau.bumb_l_b.qos_tool.loader.EpmlEpcLoader;
import de.fh_zwickau.bumb_l_b.qos_tool.loader.IEpcLoader;
import de.fh_zwickau.bumb_l_b.qos_tool.printer.IPrinter;
import de.fh_zwickau.bumb_l_b.qos_tool.printer.StaticPrinterFactory;
import de.fh_zwickau.bumb_l_b.qos_tool.printer.StaticPrinterFactory.PrinterType;
import de.fh_zwickau.bumb_l_b.qos_tool.processor.IQoSProcessor;
import de.fh_zwickau.bumb_l_b.qos_tool.processor.QoSProcessor;
import de.fh_zwickau.bumb_l_b.qos_tool.verifier.EpcVerifier;
import de.fh_zwickau.bumb_l_b.qos_tool.verifier.IEpcVerifier;


/**
 * The QoSTool. This is where everything comes together. It contains a launcher for the application.
 * Christian Fiedler
 */
public class QoSTool {
	private IPrinter printer = StaticPrinterFactory.getPrinterInstance(PrinterType.DEFAULT);

	private IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc;
	
	/**
	 * Creates a new QoSTool.
	 * @param printer Printer instance used for output.
	 */
	public QoSTool(IPrinter printer) {
		setPrinter(printer);
	}
	
	/**
	 * Executes the qos calculation.
	 */
	public void execute() {
		if(epc==null) { printer.printError(Message.get("common.no_epc_loaded")); return; }
		
		IEpcVerifier verifier = new EpcVerifier(getPrinter());
		if(!verifier.verify(epc)) { printer.printError(Message.get("common.attribute_verifiaction_failed")); return; }
		
		IQoSProcessor processor = new QoSProcessor(getPrinter());
		if(verifier.getProcessableAttributeTypes().size()==0) {
			printer.printWarning(Message.get("common.no_qos_attributes"));
			return;
		}
		for(AttributeType type : verifier.getProcessableAttributeTypes()) {
			processor.process(epc, type);
		}
	}

	/**
	 * @param printer the printer to set
	 */
	private void setPrinter(IPrinter printer) {
		if(printer==null) { throw new InvalidParameterException("printer must not be null"); }
		this.printer = printer;
	}
	
	/**
	 * Returns the current printer object.
	 * @return current printer
	 */
	public IPrinter getPrinter() {
		return printer;
	}

	/**
	 * Loads an epc stored in the given file  
	 * @param filename filename
	 */
	public void loadEpcFromFile(String filename) {
		try {
			IEpcLoader loader = new EpmlEpcLoader(); //ggf. gegen Factory ersetzen
			epc = loader.loadFile(filename);
		} catch (EpcLoaderException e) {
			printer.printError(String.format(Message.get("common.loading_file_failed"), filename), e.getElementReference());
			printer.printError(e.getMessage(), e.getElementReference());
		}
	}

}
