package de.fh_zwickau.bumb_l_b.qos_tool.customJBPT;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

/**
 * Class for managing epc attribute types
 */
public class AttributeHelper {
	/**
	 * Attribute types for qos and non-qos attribute types 
	 */
	public enum AttributeType {
		BFLOW_ID,
		COST,
		COST_MIN,
		COST_MAX,
		COST_AVG,
		TIME,
		TIME_MIN,
		TIME_MAX,
		TIME_AVG,
		RELIABILITY,
		RELIABILITY_MIN,
		RELIABILITY_MAX,
		RELIABILITY_AVG,
		PROBABILITY, //einzelne Ausführwahrscheinlichkeiten
		N_OUT_OF_M,
		PROBABILITY_LIST, //Tupel mit Ausführwahrscheinlichkeiten, nur für Schleifen
	}
	
	/** Defines supported attributes for functions */
	private static final Set<AttributeType> QOS_FUNCTION_ATTRIBUTES = new HashSet<AttributeType>(Arrays.asList(AttributeType.TIME, AttributeType.COST, AttributeType.RELIABILITY));
	/** Defines supported attributes for events */
	private static final Set<AttributeType> QOS_EVENT_ATTRIBUTES = new HashSet<AttributeType>();
	/** Defines supported attributes for edges */
	private static final Set<AttributeType> QOS_CONTROLFLOW_ATTRIBUTES = new HashSet<AttributeType>(Arrays.asList(AttributeType.PROBABILITY, AttributeType.PROBABILITY_LIST));
	/** Defines supported attributes for and connectors */
	private static final Set<AttributeType> QOS_ANDCONNECTOR_ATTRIBUTES = new HashSet<AttributeType>(Arrays.asList(AttributeType.N_OUT_OF_M));
	/** Defines supported attributes for or connectors */
	private static final Set<AttributeType> QOS_ORCONNECTOR_ATTRIBUTES = new HashSet<AttributeType>();
	/** Defines supported attributes for xor connectors */
	private static final Set<AttributeType> QOS_XORCONNECTOR_ATTRIBUTES = new HashSet<AttributeType>();
	
	/**
	 * Returns whether an attribute type can be used for qos processing in the given epc element or not
	 * @param type attribute type
	 * @param epcElement epc element
	 * @return true: attribute is suitable for the element, false: else
	 */
	public static boolean isQosAttributeForElement(AttributeType type, IContainAttribute epcElement) {
		if(epcElement instanceof CustomFunction) { return QOS_FUNCTION_ATTRIBUTES.contains(type); }
		if(epcElement instanceof CustomEvent) { return QOS_EVENT_ATTRIBUTES.contains(type); }
		if(epcElement instanceof CustomControlFlow) { return QOS_CONTROLFLOW_ATTRIBUTES.contains(type); }
		if(epcElement instanceof CustomAndConnector) { return QOS_ANDCONNECTOR_ATTRIBUTES.contains(type); }
		if(epcElement instanceof CustomOrConnector) { return QOS_ORCONNECTOR_ATTRIBUTES.contains(type); }
		if(epcElement instanceof CustomXorConnector) { return QOS_XORCONNECTOR_ATTRIBUTES.contains(type); }
		return false;
	}
	
	/**
	 * Returns whether an object is castable to a proper type for this attribute type or not.
	 * @param type
	 * @param obj
	 * @return true: type of obj is suitable, false: else
	 */
	public static boolean isAttributeOfType(AttributeType type, Object obj) {
		boolean isOfType = false;
		
		switch (type) {
			// double attribute types
			case COST:
			case COST_MIN:
			case COST_MAX:
			case COST_AVG:
			case PROBABILITY:
			case RELIABILITY:
			case RELIABILITY_MIN:
			case RELIABILITY_MAX:
			case RELIABILITY_AVG:
			case TIME:
			case TIME_MIN:
			case TIME_MAX:
			case TIME_AVG:
				isOfType = obj instanceof Double;
				break;
			// integer attribute types
			case N_OUT_OF_M:
				isOfType = obj instanceof Integer;
				break;
			//Listen
			case PROBABILITY_LIST:
				isOfType = obj instanceof ArrayList;
				break;
			// by default type should be string
			default:
				isOfType = obj instanceof String;
				break;
		}
		
		return isOfType;
	}
	
}
