package de.fh_zwickau.bumb_l_b.qos_tool.loader;

import java.util.ArrayList;
import java.util.HashMap;

import org.jbpt.graph.abs.AbstractDirectedGraph;
import org.jbpt.pm.ControlFlow;

import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.*;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper.AttributeType;

import org.jbpt.pm.FlowNode;
import org.jbpt.pm.NonFlowNode;
import org.jbpt.pm.epc.Epc;
import org.jbpt.pm.epc.IEpc;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

/**
 * Parser creating EPC models out of an EPML file.
 * 
 * Main methods: getFirstModel() and getNextModel()
 * 
 * @author gero.decker, matthias.weidlich, Anne Fischer, Christian Fiedler
 */
public class EpmlParser {
	/** Maps the epml attribute names to attribute types */
	private static final HashMap<String, AttributeType> ATTRIBUTE_MAP = new HashMap<String, AttributeType>();
	static {
		ATTRIBUTE_MAP.put("cost", AttributeType.COST);
		ATTRIBUTE_MAP.put("time", AttributeType.TIME);
		ATTRIBUTE_MAP.put("reliability", AttributeType.RELIABILITY);
		ATTRIBUTE_MAP.put("probability", AttributeType.PROBABILITY); //PROBABILITY_LIST if a probability tuple is given
		ATTRIBUTE_MAP.put("n-out-of-m", AttributeType.N_OUT_OF_M);
		ATTRIBUTE_MAP.put("N-out-of-M", AttributeType.N_OUT_OF_M);
	}
	
	protected Document doc;
	protected Node current;

	public EpmlParser(Document doc) {
		this.doc = doc;
	}

	/**
	 * Get the first model from the EPML file.
	 * 
	 * @return the epc model
	 */
	public IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> getFirstModel() {
		Node root = doc.getDocumentElement();
		if (root == null) { return null; }
		if (!root.getNodeName().toLowerCase().endsWith("epml")) { return null; } //no epml file

		current = root.getFirstChild();
		while (current != null && (!current.getNodeName().toLowerCase().endsWith("epc"))) {
			if (current.getNodeName().equals("directory")) {
				current = current.getFirstChild();
			}
			else { current = current.getNextSibling(); }
		}

		return getNextModel(); //current points to the first epc tag for every getNextModel()
	}

	/**
	 * Get the subsequent EPC model from the EPML file.
	 * 
	 * Pointer to the current model is hold in protected member "current".
	 * 
	 * @return the epc model
	 */
	public IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> getNextModel() {
		//process content of epc tags
		if (current == null || !current.getNodeName().toLowerCase().endsWith("epc")) { return null; }

		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> model = new Epc();
		model.setId(current.getAttributes().getNamedItem("epcId").getNodeValue());
		model.setName(current.getAttributes().getNamedItem("name").getNodeValue());

		addNodesAndEdges(model);

		current = current.getNextSibling();
		while (current != null && current instanceof Text) {
			current = current.getNextSibling();
		}
		//current points to the next "real" sibling
		
		return model;
	}

	/**
	 * Adds epc elements to the given model, obtained from the current node.  
	 * @param model
	 */
	protected void addNodesAndEdges(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> model) {
		addNodes(model, current.getFirstChild()); //start on first epc element
		addEdges(model, current.getFirstChild());
	}

	/**
	 * Adds nodes to the given model, obtained from the given node.
	 * @param model epc model to add elements
	 * @param node start at this node
	 */
	private void addNodes(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> model, Node node) {
		//current node reference is stored in node
		//obtain nodes
		while (getNextNonTextNode(node) != null) {
			String type = node.getNodeName();
			if (type.equals("event")) { addEvent(model, node); }
			else if (type.equals("function")) { addFunction(model, node); }
			else if (type.equals("xor")) { addXOR(model, node); } 
			else if (type.equals("and")) { addAND(model, node); } 
			else if (type.equals("or")) { addOR(model, node); }
			node = node.getNextSibling();
		}
	}
	
	/**
	 * Adds edges to the given model, obtained from the given node.
	 * @param model epc model to add elements
	 * @param node start at this node
	 */
	private void addEdges(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> model, Node node) {
		//current node reference is stored in node
		//obtain nodes
		while (getNextNonTextNode(node) != null) {
			String type = node.getNodeName();
			if (type.equals("arc")) { addArc(model, node); }
			node = node.getNextSibling();
		}
	}
	
	/**
	 * Returns the next node that is not of type Text. The given node is returned if it is not of type text.
	 * @param node start at this node
	 * @return next non-text node (including the start node)
	 */
	private Node getNextNonTextNode(Node node) {
		while(node != null && node instanceof Text) {
			node = node.getNextSibling();
		}
		return node;
	}	
	
	/**
	 * Adds an event object to the epc. The object is obtained by the given epml node.  
	 * @param model
	 * @param node
	 */
	protected void addEvent(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> model, Node node) {
		CustomEvent n = new CustomEvent();
		n.setId(getId(node));
		fetchBflowId(n,node);
		n.setName(getName(node));
		fetchAttributes(n, node);
		model.addFlowNode(n);
	}

	/**
	 * Adds a function object to the epc. The object is obtained by the given epml node.  
	 * @param model
	 * @param node
	 */
	protected void addFunction(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> model, Node node) {
		CustomFunction n = new CustomFunction();
		n.setId(getId(node));
		fetchBflowId(n,node);
		n.setName(getName(node));
		fetchAttributes(n, node);
		model.addFlowNode(n);
	}
	
	/**
	 * Adds a XOR object to the epc. The object is obtained by the given epml node.  
	 * @param model
	 * @param node
	 */
	protected void addXOR(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> model, Node node) {
		CustomXorConnector n = new CustomXorConnector();
		n.setId(getId(node));
		fetchBflowId(n,node);
		model.addFlowNode(n);
	}

	/**
	 * Adds a AND object to the epc. The object is obtained by the given epml node.  
	 * @param model
	 * @param node
	 */
	protected void addAND(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> model, Node node) {
		CustomAndConnector n = new CustomAndConnector();
		n.setId(getId(node));
		fetchBflowId(n,node);
		fetchAttributes(n, node);
		model.addFlowNode(n);
	}

	/**
	 * Adds a OR object to the epc. The object is obtained by the given epml node.  
	 * @param model
	 * @param node
	 */
	protected void addOR(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> model, Node node) {
		CustomOrConnector n = new CustomOrConnector();
		n.setId(getId(node));
		fetchBflowId(n,node);
		model.addFlowNode(n);
	}

	/**
	 * Adds a control flow edge to the epc. The object is obtained by the given epml node.
	 * @param model
	 * @param node
	 */
	protected void addArc(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> model, Node node) {
		Node flow = getChild(node, "flow");

		FlowNode source = findFlowNode(model, flow.getAttributes().getNamedItem("source").getNodeValue());
		FlowNode target = findFlowNode(model, flow.getAttributes().getNamedItem("target").getNodeValue());
//		model.addControlFlow(source, target, (Float) edgeStub.getAttribute("probability")); //alternative way
//		new ControlFlow((AbstractDirectedGraph) model, source, target, (Float) edgeStub.getAttribute("probability"));
		IContainAttribute cf = new CustomControlFlow((AbstractDirectedGraph) model, source, target);
		fetchAttributes(cf, node);
	}

	/**
	 * Returns the id-attribute from a given epml node. Returns an empty String if id is not available.
	 * @param node
	 * @return id
	 */
	protected String getId(Node node) {
		return ((node.getAttributes().getNamedItem("id")!=null) ? node.getAttributes().getNamedItem("id").getNodeValue() : "" );
	}
	
	/**
	 * Obtain the bflow id for a given epc element from the corresponding epml node. If the node does not contain any bflow id, "" is added instead.
	 * @param epcElement
	 * @param node
	 */
	protected void fetchBflowId(IContainAttribute epcElement, Node node) {
		if(node.getAttributes().getNamedItem("IdBflow") != null) {
			epcElement.addAttribute(AttributeType.BFLOW_ID, node.getAttributes().getNamedItem("IdBflow").getNodeValue());			
		}
		else {
			epcElement.addAttribute(AttributeType.BFLOW_ID, "");
		}
	}

	/**
	 * Reads out attribute tags from within the epc elements corresponding node. Supported attributes (given as list) 
	 * will be added to the epcElement.
	 * @param epcElement epc element to add the attributes
	 * @param node epml node corresponding to the epc element
	 * @param supportedAttributes list of supported attribute names
	 * @throws IllegalEpcAttributeFormatException 
	 */
	protected void fetchAttributes(IContainAttribute epcElement, Node node) throws IllegalEpcAttributeFormatException {
		Node innerNode = node.getFirstChild();
		while(innerNode != null) {
			if("attribute".equals(innerNode.getNodeName())) { //attribute tag found
				String attributeName = innerNode.getAttributes().getNamedItem("typeRef").getNodeValue();
				String attributeValue = innerNode.getAttributes().getNamedItem("value").getNodeValue();
				parseAttributeforElement(epcElement, attributeName, attributeValue);
			}			
			innerNode = innerNode.getNextSibling();
		}
	}

	/**
	 * Try to parse the attribute with the given name and value string into the given epc element.
	 * Only supported attributes are parsed. 
	 * @param epcElement element to add the attribute
	 * @param name attribute name from epml
	 * @param value attribute value from epml
	 */
	private void parseAttributeforElement(IContainAttribute epcElement, String name, String value) {
		if(ATTRIBUTE_MAP.containsKey(name)) {
			AttributeType type = ATTRIBUTE_MAP.get(name);
			if(AttributeHelper.isQosAttributeForElement(type, epcElement)) { //attribute type supported?
				//cast value into a proper type
				switch(type) {
					case TIME:
					case COST:
					case RELIABILITY:
						parseAttributeAsDouble(epcElement, type, name, value);
						break;
					case PROBABILITY:
						parseAttributeAsFlowProbability(epcElement, type, name, value);
						break;
					case N_OUT_OF_M:
						parseAttributeAsInteger(epcElement, type, name, value);
						break;
					default:
						//Strings
						epcElement.addAttribute(type, value);
				}
				
			}
		}
	}
	
	/**
	 * Try to parse the given attribute value as Double. If succesfully parsed, the value will be written as attribute of the given type
	 * to the epc element.
	 * @param epcElement element to write attribute to
	 * @param type type of attribute
	 * @param value value of attribute from epml
	 */
	private void parseAttributeAsDouble(IContainAttribute epcElement, AttributeType type, String name, String value) {
		value = value.replace(",", "."); //comma to point
		try {
			epcElement.addAttribute(type, new Double(value));
		} catch (NumberFormatException e) {
			epcElement.addAttribute(type, Double.NaN); //Parser tolerates non numeric values as verifier will recognize and report this
//			throw new IllegalEpcAttributeFormatException("'"+value+"' is no proper value for attribute "+name+" of type "+type, ""+epcElement.getAttribute(AttributeType.BFLOW_ID));
		}
	}
	
	/**
	 * Try to parse the given attribute value as Integer. If succesfully parsed, the value will be written as attribute of the given type
	 * to the epc element.
	 * @param epcElement element to write attribute to
	 * @param type type of attribute
	 * @param value value of attribute from epml
	 */
	private void parseAttributeAsInteger(IContainAttribute epcElement, AttributeType type, String name, String value) {
		try {
			epcElement.addAttribute(type, new Integer(value));
		} catch (NumberFormatException e) {
			epcElement.addAttribute(type, Integer.MIN_VALUE); //Parser tolerates non numeric values as verifier will recognize and report this
//			throw new IllegalEpcAttributeFormatException("'"+value+"' is no proper value for attribute "+name+" of type "+type, ""+epcElement.getAttribute(AttributeType.BFLOW_ID));
		}
	}
	
	/**
	 * Try to parse the given attribute value as Double. If succesfully parsed, the value will be written as attribute of the given type
	 * to the epc element.
	 * @param epcElement element to write attribute to
	 * @param type type of attribute
	 * @param value value of attribute from epml
	 */
	private void parseAttributeAsFlowProbability(IContainAttribute epcElement, AttributeType type, String name, String value) {
		// probability="0.5" or probability="(0.5 0.2 0.1)"
		// decide what is given
		if(value.matches("\\s*\\(\\s*(.*)\\s*\\)\\s*")) {
			//tuple
			type = AttributeType.PROBABILITY_LIST; //internally use this attribute for probability tuples
			value = value.replaceAll("\\s*\\(\\s*(.*)\\s*\\)\\s*", "$1"); //remove brackets and whitespaces
			String[] attributeValueParts = value.split("\\s+");
			ArrayList<Double> probabilityList = new ArrayList<Double>();
			for(String attrValuePart : attributeValueParts) {
				if(!"".equals(attrValuePart)) {
					try {
						attrValuePart = attrValuePart.replace(",", "."); //comma to point
						probabilityList.add(new Double(attrValuePart));
					} catch (NumberFormatException e) {
						probabilityList.add(Double.NaN); //Parser tolerates non numeric values as verifier will recognize and report this
	//					throw new IllegalEpcAttributeFormatException("'"+value+"' is no proper value for attribute "+name+" of type "+type, ""+epcElement.getAttribute(AttributeType.BFLOW_ID));
					}
				}
			}
			//Parser tolerates empty probability lists as verifier will recognize an report this
//			if(probabilityList.size()==0) {
//				throw new IllegalEpcAttributeFormatException("Attribute "+name+" of type "+type+" must not be empty", ""+epcElement.getAttribute(AttributeType.BFLOW_ID));
//			}
			epcElement.addAttribute(type, probabilityList);
		}
		else {
			//single value
			parseAttributeAsDouble(epcElement, type, name, value);
		}
	}
	
	/**
	 * Returns the name for an epml node. Critical characters are replaced.
	 * @param node
	 * @return
	 */
	protected String getName(Node node) {
		String name = getChild(node, "name").getTextContent();

//		name = name.toLowerCase();
		name = name.replace("-", " ");
		name = name.replace(".", "");
		name = name.replace("(", "");
		name = name.replace(")", "");
		name = name.replace("[", "");
		name = name.replace("]", "");
		name = name.replace(",", " ");
		name = name.replace(";", " ");
		name = name.replace("\n", " ");
		name = name.replace("\r", " ");
		name = name.replace("\t", " ");
		name = name.replace("\\n", " ");
		name = name.replace("\\r", " ");
		name = name.replace("\\t", " ");
		name = name.replace("\\", " ");
		return name;
	}

	/**
	 * Returns the first child node of a given node n which has the given name name.
	 * @param n
	 * @param name
	 * @return
	 */
	protected Node getChild(Node n, String name) {
		for (Node node = n.getFirstChild(); node != null; node = node.getNextSibling())
			if (node.getNodeName().indexOf(name) >= 0) { return node; }
		return null;
	}

	/**
	 * Searches the given model for a FlowNode with the given id. Returns the FlowNode corresponding to that id. If no such node
	 * is found, null is returned.
	 * @param model
	 * @param id
	 * @return
	 */
	//TODO: Methode auslagern, das gehört nicht wirklich zum Parser sondern eher in eine statische Toolklasse für JBPT!
	protected FlowNode findFlowNode(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> model, String id) {
		for (FlowNode n : model.getFlowNodes()) {
			if (n.getId().equals(id))
				return n;
		}
		return null;
	}

}
