package de.fh_zwickau.bumb_l_b.qos_tool.processor;

import java.util.Arrays;

/**
 * A class to compute combinations.
 */
public class Combinator {
	private int[] combi;
	private int n;
	private int k;
	private int combiCount;
	
	/**
	 * Creates a new Combinator which generates combinations for n choose k
	 * @param n number of elements
	 * @param k number of elements to choose
	 */
	public Combinator(int n, int k) {
		if(n<1) { throw new IllegalArgumentException("n must not be lower than 1"); }
		if(n<k) { throw new IllegalArgumentException("n must not be lower than k"); }
		if(k<1) { throw new IllegalArgumentException("k must not be lower than 1"); }
		this.n = n;
		this.k = k;
		combiCount = binomial(n, k);
		reset();
	}
	
	/**
	 * Resets the combinator to the initial combination
	 */
	public void reset() {
		combi = new int[k];
		for(int i=0; i<k; i++) {
			combi[i] = i;
		}
	}
	
	/**
	 * Compute the next combination.
	 * @return true: next combination computed, false: no further combination available
	 */
	public boolean next() {
		int[] tmp = Arrays.copyOf(combi, combi.length);
		if(!increase(k-1)) {
			combi = tmp; //restore old value
			return false;
		}
		return true;
		
	}
	
	/**
	 * Increases the combination-value at position pos.
	 * @param pos
	 * @return
	 */
	private boolean increase(int pos) {
		if(pos < 0 || pos>=k) { return false; }
		combi[pos] += 1;
		if(combi[pos] >= (n-k+(pos+1))) { //overflow
			if(increase(pos-1)) { //carry to next position
				combi[pos] = combi[pos-1]+1;
			}
			else {
				return false; //last position reached
			}
		}
		return true;
	}
	
	/**
	 * Returns the current combination.
	 * @return
	 */
	public int[] getCombination() {
		return combi;
	}
	
	/**
	 * Returns the total number of combinations.
	 * @return
	 */
	public int getCombinationCount() {
		return combiCount;
	}
	
	/**
	 * Computes the number of combinations for n choose k
	 * @param n number of elements
	 * @param k choose k elements
	 * @return number of combinations
	 */
	private int binomial(final int n, final int k) {
	    long ret = 1;
	    for (int ik = 0; ik < k; ik++) {
	        ret = ret * (n-ik) / (ik+1);
	        if(ret < 1) { throw new ArithmeticException("Number overflow while calculating combinations"); }
	    }
	    if(ret > Integer.MAX_VALUE) { throw new ArithmeticException("Maximum number of combinations exceeded."); }
	    return (int)ret;
	}
}
