package de.fh_zwickau.bumb_l_b.qos_tool.processor;

import org.jbpt.graph.abs.AbstractDirectedGraph;
import org.jbpt.pm.ControlFlow;
import org.jbpt.pm.FlowNode;
import org.jbpt.pm.NonFlowNode;
import org.jbpt.pm.epc.Epc;
import org.jbpt.pm.epc.IEpc;

import de.fh_zwickau.bumb_l_b.qos_tool.Message;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper.AttributeType;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomControlFlow;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.IContainAttribute;
import de.fh_zwickau.bumb_l_b.qos_tool.printer.IPrinter;

/**
 * A qos processor that uses PatternMatcher and Reducer to compute QoS values
 */
public class QoSProcessor implements IQoSProcessor {
	private IPrinter printer;

	public QoSProcessor(IPrinter printer) {
		this.printer = printer;
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public void process(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc, AttributeType type) {
		Reducer reducer = null;
		AttributeType resultMin, resultMax, resultAvg;
		
		switch(type) {
			case COST:
				reducer = new Reducer((IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode>) cloneEpc(epc), new CostReplacementStrategy());
				resultMin = AttributeType.COST_MIN;
				resultMax = AttributeType.COST_MAX;
				resultAvg = AttributeType.COST_AVG;
				break;
			case TIME:
				reducer = new Reducer((IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode>) cloneEpc(epc), new TimeReplacementStrategy());
				resultMin = AttributeType.TIME_MIN;
				resultMax = AttributeType.TIME_MAX;
				resultAvg = AttributeType.TIME_AVG;
				break;
			case RELIABILITY:
				reducer = new Reducer((IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode>) cloneEpc(epc), new ReliabilityReplacementStrategy());
				resultMin = AttributeType.RELIABILITY_MIN;
				resultMax = AttributeType.RELIABILITY_MAX;
				resultAvg = AttributeType.RELIABILITY_AVG;
				break;
			default:
				printer.printError(String.format(Message.get("processor.unsupported_attribute_type"), type));
				return;
		}
		
		//try to compute min, max, avg using the current reducer
		try {
			if(reducer.reduce()) {
				for(FlowNode node : reducer.getEpc().getFlowNodes()) {
					IContainAttribute tmp = (IContainAttribute) node;
					printer.printInfo(String.format(Message.get("processor.result"), type, "min", roundTo4SignificantDigits((Double) tmp.getAttribute(resultMin))));
					printer.printInfo(String.format(Message.get("processor.result"), type, "max", roundTo4SignificantDigits((Double) tmp.getAttribute(resultMax))));
					printer.printInfo(String.format(Message.get("processor.result"), type, "avg", roundTo4SignificantDigits((Double) tmp.getAttribute(resultAvg))));
				}
			}
			else {
				printer.printError(Message.get("processor.calculation_failed"));
			}
		} catch (ReplacementException e) {
			if(e.getAffectedNode() != null && e.getAffectedNode() instanceof IContainAttribute) {
				printer.printError(e.getLocalizedMessage(), ((IContainAttribute) e.getAffectedNode()).getAttribute(AttributeType.BFLOW_ID).toString());				
			}
			else {
				printer.printError(e.getLocalizedMessage());				
			}
		}

	}

	/**
	 * round a double to the 4 significant digits
	 * @param num
	 * @return
	 */
	public static double roundTo4SignificantDigits(double num) {
	    if(num == 0) {
	        return 0;
	    }

	    if(num == Double.POSITIVE_INFINITY) {
	    	return Double.POSITIVE_INFINITY;
	    }
	    
	    final double d = Math.ceil(Math.log10(num < 0 ? -num: num));
	    final int power = 4 - (int) d;

	    final double magnitude = Math.pow(10, power);
	    final long shifted = Math.round(num*magnitude);
	    
	    return shifted/magnitude;
	}
	
	/**
	 * Create a clone of the given epc. The nodes in the clone are the same as in the blueprint. Whereas edges (control flows) are
	 * newly created.
	 * @param epc blueprint
	 * @return clone
	 */
	public static IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> cloneEpc(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc) {
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> clone = new Epc();
		for(FlowNode node : epc.getVertices()) {
			clone.addVertex(node);
		}
		for(ControlFlow<FlowNode> edge : epc.getEdges()) {
			IContainAttribute newEdge = new CustomControlFlow((AbstractDirectedGraph) clone, edge.getSource(), edge.getTarget());
			if(edge instanceof IContainAttribute) {
				IContainAttribute tmp = ((IContainAttribute)edge);
				for(AttributeType key : tmp.getAttributeKeys()) {
					newEdge.addAttribute(key, tmp.getAttribute(key));
				}
			}
		}
		return clone;
	}
	
}
