package de.fh_zwickau.bumb_l_b.qos_tool.processor;

import org.jbpt.graph.abs.AbstractDirectedGraph;
import org.jbpt.pm.ControlFlow;
import org.jbpt.pm.FlowNode;
import org.jbpt.pm.NonFlowNode;
import org.jbpt.pm.epc.IEpc;

import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper.AttributeType;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomControlFlow;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomEvent;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomFunction;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.IContainAttribute;
import de.fh_zwickau.bumb_l_b.qos_tool.processor.pattern.AndPatternMatcher;
import de.fh_zwickau.bumb_l_b.qos_tool.processor.pattern.CyclePatternMatcher;
import de.fh_zwickau.bumb_l_b.qos_tool.processor.pattern.EmptyPatternMatcher;
import de.fh_zwickau.bumb_l_b.qos_tool.processor.pattern.LinearPatternMatcher;
import de.fh_zwickau.bumb_l_b.qos_tool.processor.pattern.PatternMatcher;
import de.fh_zwickau.bumb_l_b.qos_tool.processor.pattern.UselessGatewayPatternMatcher;
import de.fh_zwickau.bumb_l_b.qos_tool.processor.pattern.XorPatternMatcher;
import de.fh_zwickau.bumb_l_b.qos_tool.processor.pattern.PatternMatcher.PatternMatcherResult;

/**
 * A class that perform pattern search and graph reduction to reduce epcs to a single element. 
 */
public class Reducer {
	private IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc;
	private IReplacementStrategy replacementStrategy;
	private PatternMatcher uselessGwMatcher;
	private PatternMatcher emptyMatcher;
	private PatternMatcher linearMatcher;
	private PatternMatcher andMatcher;
	private PatternMatcher xorMatcher;
	private PatternMatcher cycleMatcher;
	
	/**
	 * Create a new instance.
	 * @param epc
	 */
	public Reducer(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc, IReplacementStrategy replacementStrategy) {
		if(epc == null) { throw new IllegalArgumentException("epc must not be null"); }
		if(replacementStrategy == null) { throw new IllegalArgumentException("replacementStrategy must not be null"); }
		this.epc = epc;
		this.replacementStrategy = replacementStrategy;
		uselessGwMatcher = new UselessGatewayPatternMatcher(this.epc);
		emptyMatcher = new EmptyPatternMatcher(this.epc);
		linearMatcher = new LinearPatternMatcher(this.epc);
		andMatcher = new AndPatternMatcher(this.epc);
		xorMatcher = new XorPatternMatcher(this.epc);
		cycleMatcher = new CyclePatternMatcher(this.epc);
	}
	
	/**
	 * Perform the reduction.
	 * @return true: reduction successful; false: reduction failed
	 * @throws ReplacementException 
	 */
	public boolean reduce() throws ReplacementException {
		if(epc.getVertices().size()<1) { return false; } //empty epc
		
//		System.out.println("epc size beginning "+epc.getVertices().size());
		fillEmpty();
//		System.out.println("epc size after empty "+epc.getVertices().size());
		removeUselessGateways();
//		System.out.println("epc size after uselessGw "+epc.getVertices().size());
		addMissingAttributeValues();
		
		int noReduction = 0;
		boolean reduction = false;
		while(epc.getVertices().size() > 1) {
			//always reduce linear
			reduction = replaceByPattern(linearMatcher);
//			System.out.println("epc size after linear "+epc.getVertices().size());
			if(epc.getVertices().size() == 1) { break; }
			
			//reduce other patterns; start next turn if one matched
			if(replaceByPattern(andMatcher)) { continue; } //next turn
			if(replaceByPattern(xorMatcher)) { continue; } //next turn
			if(replaceByPattern(cycleMatcher)) { continue; } //next turn
			
			//nothing reduced in this turn
			if(!reduction) {
				noReduction++;
				if(noReduction >= 2) { return false; } //cannot reduce chain with available patterns
			}
		}
//		System.out.println("epc size finally "+epc.getVertices().size());
		return true;
	}

	/**
	 * Add missing attribute values to function and events. Usually the values should be neutral i.e. identity elements for calculations.
	 */
	private void addMissingAttributeValues() {
		for(FlowNode node : epc.getFlowNodes()) {
			if(node instanceof CustomFunction || node instanceof CustomEvent) {
				if(!(node instanceof IContainAttribute)) { throw new RuntimeException("node should have been of type IContainAttribute"); }
				replacementStrategy.addMissingAttributeValues((IContainAttribute)node);
			}
		}
	}

	/**
	 * Add dummy nodes between gateways which are direct neighbors.
	 * @throws ReplacementException 
	 */
	private void fillEmpty() throws ReplacementException {
		emptyMatcher.find();
		for(PatternMatcherResult result : emptyMatcher.getResults()) {
			//obtain a replacement
			FlowNode replacement = replacementStrategy.replaceEmpty(result);
			//replace
			ControlFlow<FlowNode> edgeToRemove = result.getEntryEdge();
			epc.addFlowNode(replacement);
			addControlFlow(result.getPreEntry(), replacement, (IContainAttribute) edgeToRemove);
			addControlFlow(replacement, result.getPostExit());
			epc.removeEdge(edgeToRemove);
		}
	}
	
	/**
	 * Remove useless Gateways.
	 */
	private void removeUselessGateways() {
		uselessGwMatcher.find();
		for(PatternMatcherResult result : uselessGwMatcher.getResults()) {
			if(result.getPreEntry() != null && result.getPostExit() != null) {
				addControlFlow(result.getPreEntry(), result.getPostExit(), (IContainAttribute) result.getEntryEdge());				
			}
			epc.removeVertices(result.getNodes());			
		}
	}
	
	/**
	 * Replaces regions that are matched by the given pattern matcher by single nodes.
	 * @param pattern pattern matcher to use
	 * @return true: epc size reduced; false: epc size not reduced
	 * @throws ReplacementException 
	 */
	private boolean replaceByPattern(PatternMatcher pattern) throws ReplacementException {
		int oldsize = epc.getVertices().size();
		pattern.find();
		for(PatternMatcherResult result : pattern.getResults()) {
			//obtain replacement
			FlowNode replacement;
			if(pattern instanceof LinearPatternMatcher) { replacement = replacementStrategy.replaceLinear(result); }
			else if(pattern instanceof AndPatternMatcher) { replacement = replacementStrategy.replaceAnd(result); }
			else if(pattern instanceof XorPatternMatcher) { replacement = replacementStrategy.replaceXor(result); }
			else if(pattern instanceof CyclePatternMatcher) { replacement = replacementStrategy.replaceCycle(result); }
			else { throw new IllegalArgumentException("Unknown pattern for this method."); }
			
//			//helpful for debugging fuzzy math
//			System.out.print(result+" values : ");
//			for(AttributeType key : ((IContainAttribute)replacement).getAttributeKeys()) {
//				System.out.print(key+"="+((IContainAttribute)replacement).getAttribute(key)+" ");
//			}
//			System.out.println();
			
			//replace
			epc.addFlowNode(replacement);
			addControlFlow(result.getPreEntry(), replacement, (IContainAttribute) result.getEntryEdge());
			addControlFlow(replacement, result.getPostExit());
			epc.removeVertices(result.getNodes());
		}
		return oldsize > epc.getVertices().size();
	}
	
	/**
	 * Creates a new ControlFlow of type CustomControlFlow (!) within the current epc.
	 * @param from source node
	 * @param to target node
	 * @return the new control flow
	 */
	private IContainAttribute addControlFlow(FlowNode from, FlowNode to) {
		return addControlFlow(from, to, null);
	}
	
	/**
	 * Creates a new ControlFlow of type CustomControlFlow (!) within the current epc. Copy all attribute values from
	 * attributeSource to the new CustomControlFlow
	 * @param from source node
	 * @param to target node
	 * @param attributeSource
	 * @return the new control flow
	 */
	private IContainAttribute addControlFlow(FlowNode from, FlowNode to, IContainAttribute attributeSource) {
		IContainAttribute edge = new CustomControlFlow((AbstractDirectedGraph) epc, from, to);
//		System.out.println("new edge "+edge+" with attribute Source "+attributeSource);
		if(attributeSource != null) {
			for(AttributeType key : attributeSource.getAttributeKeys()) {
//				System.out.println("copy "+key+" from "+attributeSource+" to "+edge);
				edge.addAttribute(key, attributeSource.getAttribute(key));
			}
		}
		
		return edge;
	}
	
	/**
	 * Returns the epc in it current (maybe reduced) state.
	 * @return
	 */
	public IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> getEpc() {
		return epc;
	}
}
