package de.fh_zwickau.bumb_l_b.qos_tool.processor;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.jbpt.pm.ControlFlow;
import org.jbpt.pm.FlowNode;
import org.jbpt.pm.NonFlowNode;
import org.jbpt.pm.epc.IEpc;

import de.fh_zwickau.bumb_l_b.qos_tool.Message;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper.AttributeType;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomEvent;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomFunction;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.IContainAttribute;
import de.fh_zwickau.bumb_l_b.qos_tool.processor.pattern.PatternMatcher.PatternMatcherResult;

public class ReliabilityReplacementStrategy implements IReplacementStrategy {

	/**
	 * {@inheritDoc}
	 */
	@Override
	public FlowNode replaceEmpty(final PatternMatcherResult result) {
		IContainAttribute replacement = new CustomFunction("Dummy");
		addMissingAttributeValues(replacement);
		return (FlowNode)replacement;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public FlowNode replaceLinear(final PatternMatcherResult result) {
		//create new node and set and fetch default values
		IContainAttribute replacement = new CustomFunction(result.getEntry().getLabel()+"~"+result.getExit().getLabel());
		addMissingAttributeValues(replacement);
		double min = (double) replacement.getAttribute(AttributeType.RELIABILITY_MIN);
		double max = (double) replacement.getAttribute(AttributeType.RELIABILITY_MAX);
		double avg = (double) replacement.getAttribute(AttributeType.RELIABILITY_AVG);
		
		//in linear chains we just have to multiply
		for(FlowNode node : result.getNodes()) {
			if(node instanceof CustomFunction || node instanceof CustomEvent) {
				IContainAttribute tmp = (IContainAttribute) node;
				min *= (Double) tmp.getAttribute(AttributeType.RELIABILITY_MIN);
				max *= (Double) tmp.getAttribute(AttributeType.RELIABILITY_MAX);
				avg *= (Double) tmp.getAttribute(AttributeType.RELIABILITY_AVG);
			}
		}
		
		//write values to the new element
		replacement.setAttribute(AttributeType.RELIABILITY_MIN, min);
		replacement.setAttribute(AttributeType.RELIABILITY_MAX, max);
		replacement.setAttribute(AttributeType.RELIABILITY_AVG, avg);
		
		return (FlowNode)replacement;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public FlowNode replaceAnd(final PatternMatcherResult result) {
		//create new node and set and fetch default values
		IContainAttribute replacement = new CustomFunction(result.getEntry().getLabel()+"~"+result.getExit().getLabel());
		addMissingAttributeValues(replacement);
		double min = 1;
		double max = 1;
		double avg = 0;
		
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = result.getOwner().getEpc();
		
		//fetch n-out-of-m value from join (if not given, all paths must finish thus n=m)
		//n: number of paths that must finish; m: number of all paths in and region
		int m = epc.getIncomingEdges(result.getExit()).size();
		int n = m;
		IContainAttribute join = ((IContainAttribute)result.getExit());
		if(join.hasAttribute(AttributeType.N_OUT_OF_M)) {
			n = (int) join.getAttribute(AttributeType.N_OUT_OF_M);
		}
		
		//get the slowest out of the n fastest paths for min,max,avg
		ArrayList<IContainAttribute> nodes = new ArrayList<IContainAttribute>();
		ArrayList<Double> minValues = new ArrayList<Double>();
		ArrayList<Double> maxValues = new ArrayList<Double>();
		ArrayList<Double> avgValues = new ArrayList<Double>();
		for(FlowNode node : epc.getDirectPredecessors(result.getExit())) {
			IContainAttribute tmp = (IContainAttribute)node;
			nodes.add(tmp); //collect all paths for later use in avg calculation
			minValues.add((Double) tmp.getAttribute(AttributeType.RELIABILITY_MIN));
			maxValues.add((Double) tmp.getAttribute(AttributeType.RELIABILITY_MAX));
			avgValues.add((Double) tmp.getAttribute(AttributeType.RELIABILITY_AVG));
		}
		Collections.sort(minValues);
		Collections.sort(maxValues);
		Collections.sort(avgValues);
		
		//min is product of n smallest reliabilities
		//max is product of n largest reliabilities
		for(int i=0; i<n; i++) {
			min *= minValues.get(i);
			max *= maxValues.get((m-i)-1);
		}
		
		//average of all possible combinations
		Combinator combinator = new Combinator(m, n);
		int numCombinations = combinator.getCombinationCount();
//		System.out.println("num combinations: "+numCombinations);		
		
		do { //iterate over all possible combinations
			int[] combi = combinator.getCombination();
//			System.out.println(Arrays.toString(combi));
			//determine avg reliability of current path
			double combiAvg = 1;
			for(int key : combi) {
				combiAvg *= (double) nodes.get(key).getAttribute(AttributeType.RELIABILITY_AVG);
			}
			avg += (combiAvg / numCombinations); //add interest to mean value
		} while(combinator.next());
		
		//write values to the new element
		replacement.setAttribute(AttributeType.RELIABILITY_MIN, min);
		replacement.setAttribute(AttributeType.RELIABILITY_MAX, max);
		replacement.setAttribute(AttributeType.RELIABILITY_AVG, avg);
		
		return (FlowNode)replacement;
	}

	/**
	 * {@inheritDoc}
	 * @throws ReplacementException 
	 */
	@Override
	public FlowNode replaceXor(final PatternMatcherResult result) throws ReplacementException {
		//create new node and set and fetch default values
		IContainAttribute replacement = new CustomFunction(result.getEntry().getLabel()+"~"+result.getExit().getLabel());
		addMissingAttributeValues(replacement);
		double min = 1;
		double max = 0;
		double avg = 0;
		
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = result.getOwner().getEpc();
		
		for(ControlFlow<FlowNode> edge : epc.getEdgesWithSource(result.getEntry())) {
			if(!((IContainAttribute)edge).hasAttribute(AttributeType.PROBABILITY)) {
				throw new ReplacementException(Message.get("verifier.probability_value_missing"), result.getEntry());
			}
			double probability = (double) ((IContainAttribute)edge).getAttribute(AttributeType.PROBABILITY);
			FlowNode node = edge.getTarget();
			min = Math.min(min, (double) ((IContainAttribute)node).getAttribute(AttributeType.RELIABILITY_MIN));
			max = Math.max(max, (double) ((IContainAttribute)node).getAttribute(AttributeType.RELIABILITY_MAX));
			avg += probability * (double) ((IContainAttribute)node).getAttribute(AttributeType.RELIABILITY_AVG);
		}
		
		//write values to the new element
		replacement.setAttribute(AttributeType.RELIABILITY_MIN, min);
		replacement.setAttribute(AttributeType.RELIABILITY_MAX, max);
		replacement.setAttribute(AttributeType.RELIABILITY_AVG, avg);
		
		return (FlowNode)replacement;
	}

	/**
	 * {@inheritDoc}
	 * @throws ReplacementException 
	 */
	@Override
	public FlowNode replaceCycle(final PatternMatcherResult result) throws ReplacementException {
		//create new node and set and fetch default values
		IContainAttribute replacement = new CustomFunction(result.getEntry().getLabel()+"~"+result.getExit().getLabel());
		addMissingAttributeValues(replacement);
		double min = Double.NaN;
		double max = Double.NaN;
		double avg = Double.NaN;
			
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = result.getOwner().getEpc();

		//follow main path and get values
		IContainAttribute mainPathNode = ((IContainAttribute)epc.getFirstDirectSuccessor(result.getEntry()));
		double minMainPath = (double) mainPathNode.getAttribute(AttributeType.RELIABILITY_MIN);
		double maxMainPath = (double) mainPathNode.getAttribute(AttributeType.RELIABILITY_MAX);
		double avgMainPath = (double) mainPathNode.getAttribute(AttributeType.RELIABILITY_AVG);
		
		//follow split's outgoing paths to find back path
		for(ControlFlow<FlowNode> edge : epc.getOutgoingEdges(result.getExit())) {
			
			if(epc.getFirstDirectSuccessor(edge.getTarget())==result.getEntry()) {
				//this is the back path
				IContainAttribute backPathNode = (IContainAttribute)edge.getTarget();
				IContainAttribute backPathEdge = (IContainAttribute) edge;
				double minBackPath = (double) backPathNode.getAttribute(AttributeType.RELIABILITY_MIN);
				double maxBackPath = (double) backPathNode.getAttribute(AttributeType.RELIABILITY_MAX);
				double avgBackPath = (double) backPathNode.getAttribute(AttributeType.RELIABILITY_AVG);
				
				max = maxMainPath; //main path is always executed (at least once)
				
				//calculate as geometric series
				if(backPathEdge.hasAttribute(AttributeType.PROBABILITY)) {
					double probability = (double) backPathEdge.getAttribute(AttributeType.PROBABILITY);
					if(Math.min(minMainPath, minBackPath)==1) { min=1; } //reliability does not change
					else { min=0; } //infinite repeating causes infinite decrease of reliability
					//a hint to understand the calculation
					//sum[from i=0 to infinity](a*p^i) for 0 < p < 1, a=const
					//= a * (1-lim[i->infinity](p^i)) / (1-p)
					//= a * (1-0) / (1-p)
					//avg = sum[from i=0 to infinity]( (1-p) * p^i * r_main * r_main^i * r_back^i )
					//    = (1-p) * r_main * sum[from i=0 to infinity]( (p*r_main*r_back)^i ) 
					avg = (1-probability) * avgMainPath * (1/(1-(probability*avgMainPath*avgBackPath)));
				}
				//calculate using tuple
				else if(backPathEdge.hasAttribute(AttributeType.PROBABILITY_LIST)) {
//					System.out.println(backPathEdge.getAttribute(AttributeType.PROBABILITY_LIST));
					List<Double> probabilities = (List<Double>) backPathEdge.getAttribute(AttributeType.PROBABILITY_LIST);
					min = minMainPath * Math.pow((minMainPath*minBackPath), probabilities.size());
					
					avg = (1-probabilities.get(0))*avgMainPath;
					double fullPathProbability = 1.0;
					for(int i=0; i<probabilities.size(); i++) {
						fullPathProbability *= probabilities.get(i); //this is the overall chance to cycle back in the n-th path
						if(i < probabilities.size()-1) {
							avg += (1-probabilities.get(i+1)) * fullPathProbability * avgMainPath * Math.pow((avgMainPath * avgBackPath), (i+1));
						}
						else {
							//in the last cycle the probability for the back path is 0
							avg += fullPathProbability * avgMainPath * Math.pow((avgMainPath * avgBackPath), (i+1));							
						}
					}
				}
				//cannot calculate without probabilities
				else {
					throw new ReplacementException(Message.get("verifier.probability_value_missing_on_cycle"), result.getEntry());
				}
				break; //only need the cyclic path, skip others
			}
			else { //on none cyclic paths perform some checks
				if(((IContainAttribute) edge).hasAttribute(AttributeType.PROBABILITY_LIST)) {
					throw new ReplacementException(Message.get("verifier.probability_list_on_wrong_edge"), result.getEntry());
				}
			}
		}
		
		//write values to the new element
		replacement.setAttribute(AttributeType.RELIABILITY_MIN, min);
		replacement.setAttribute(AttributeType.RELIABILITY_MAX, max);
		replacement.setAttribute(AttributeType.RELIABILITY_AVG, avg);
		
		return (FlowNode)replacement;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void addMissingAttributeValues(final IContainAttribute node) {
		Double reliability;
		if(node.hasAttribute(AttributeType.RELIABILITY)) {
			reliability = (Double) node.getAttribute(AttributeType.RELIABILITY);
		}
		else {
			reliability = 1.0;
			node.setAttribute(AttributeType.RELIABILITY, reliability);
		}
		//prepare min, max, avg
		node.setAttribute(AttributeType.RELIABILITY_MIN, reliability);
		node.setAttribute(AttributeType.RELIABILITY_MAX, reliability);
		node.setAttribute(AttributeType.RELIABILITY_AVG, reliability);
	}
	
}
