package de.fh_zwickau.bumb_l_b.qos_tool.processor;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.jbpt.pm.ControlFlow;
import org.jbpt.pm.FlowNode;
import org.jbpt.pm.NonFlowNode;
import org.jbpt.pm.epc.IEpc;

import de.fh_zwickau.bumb_l_b.qos_tool.Message;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper.AttributeType;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomEvent;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomFunction;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.IContainAttribute;
import de.fh_zwickau.bumb_l_b.qos_tool.processor.pattern.PatternMatcher.PatternMatcherResult;

public class TimeReplacementStrategy implements IReplacementStrategy {

	/**
	 * {@inheritDoc}
	 */
	@Override
	public FlowNode replaceEmpty(final PatternMatcherResult result) {
		IContainAttribute replacement = new CustomFunction("Dummy");
		addMissingAttributeValues(replacement);
		return (FlowNode)replacement;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public FlowNode replaceLinear(final PatternMatcherResult result) {
		//create new node and set and fetch default values
		IContainAttribute replacement = new CustomFunction(result.getEntry().getLabel()+"~"+result.getExit().getLabel());
		addMissingAttributeValues(replacement);
		double min = (double) replacement.getAttribute(AttributeType.TIME_MIN);
		double max = (double) replacement.getAttribute(AttributeType.TIME_MAX);
		double avg = (double) replacement.getAttribute(AttributeType.TIME_AVG);
		
		//in linear chains we just have to summarize
		for(FlowNode node : result.getNodes()) {
			if(node instanceof CustomFunction || node instanceof CustomEvent) {
				IContainAttribute tmp = (IContainAttribute) node;
				min += (Double) tmp.getAttribute(AttributeType.TIME_MIN);
				max += (Double) tmp.getAttribute(AttributeType.TIME_MAX);
				avg += (Double) tmp.getAttribute(AttributeType.TIME_AVG);
			}
		}
		
		//write values to the new element
		replacement.setAttribute(AttributeType.TIME_MIN, min);
		replacement.setAttribute(AttributeType.TIME_MAX, max);
		replacement.setAttribute(AttributeType.TIME_AVG, avg);
		
		return (FlowNode)replacement;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public FlowNode replaceAnd(final PatternMatcherResult result) {
		//create new node and set and fetch default values
		IContainAttribute replacement = new CustomFunction(result.getEntry().getLabel()+"~"+result.getExit().getLabel());
		addMissingAttributeValues(replacement);
		double min = Double.NaN;
		double max = Double.NaN;
		double avg = Double.NaN;
		
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = result.getOwner().getEpc();
		
		//fetch n-out-of-m value from join (if not given, all paths must finish thus n=m)
		//n: number of paths that must finish; m: number of all paths in and region
		int m = epc.getIncomingEdges(result.getExit()).size();
		int n = m;
		IContainAttribute join = ((IContainAttribute)result.getExit());
		if(join.hasAttribute(AttributeType.N_OUT_OF_M)) {
			n = (int) join.getAttribute(AttributeType.N_OUT_OF_M);
		}
		
		//get the slowest out of the n fastest paths for min,max,avg
		ArrayList<Double> minValues = new ArrayList<Double>();
		ArrayList<Double> maxValues = new ArrayList<Double>();
		ArrayList<Double> avgValues = new ArrayList<Double>();
		for(FlowNode node : epc.getDirectPredecessors(result.getExit())) {
			IContainAttribute tmp = (IContainAttribute)node;
			minValues.add((Double) tmp.getAttribute(AttributeType.TIME_MIN));
			maxValues.add((Double) tmp.getAttribute(AttributeType.TIME_MAX));
			avgValues.add((Double) tmp.getAttribute(AttributeType.TIME_AVG));
		}
		Collections.sort(minValues);
		Collections.sort(maxValues);
		Collections.sort(avgValues);
		min = minValues.get(n-1);
		max = maxValues.get(n-1);
		avg = avgValues.get(n-1);
				
		//write values to the new element
		replacement.setAttribute(AttributeType.TIME_MIN, min);
		replacement.setAttribute(AttributeType.TIME_MAX, max);
		replacement.setAttribute(AttributeType.TIME_AVG, avg);
		
		return (FlowNode)replacement;
	}

	/**
	 * {@inheritDoc}
	 * @throws ReplacementException 
	 */
	@Override
	public FlowNode replaceXor(final PatternMatcherResult result) throws ReplacementException {
		//create new node and set and fetch default values
		IContainAttribute replacement = new CustomFunction(result.getEntry().getLabel()+"~"+result.getExit().getLabel());
		addMissingAttributeValues(replacement);
		double min = Double.POSITIVE_INFINITY;
		double max = 0;
		double avg = 0;
		
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = result.getOwner().getEpc();
		
		for(ControlFlow<FlowNode> edge : epc.getEdgesWithSource(result.getEntry())) {
			if(!((IContainAttribute)edge).hasAttribute(AttributeType.PROBABILITY)) {
				throw new ReplacementException(Message.get("verifier.probability_value_missing"), result.getEntry());
			}
			double probability = (double) ((IContainAttribute)edge).getAttribute(AttributeType.PROBABILITY);
			FlowNode node = edge.getTarget();
			min = Math.min(min, (double) ((IContainAttribute)node).getAttribute(AttributeType.TIME_MIN));
			max = Math.max(max, (double) ((IContainAttribute)node).getAttribute(AttributeType.TIME_MAX));
			avg += probability * (double) ((IContainAttribute)node).getAttribute(AttributeType.TIME_AVG);
		}
		
		//write values to the new element
		replacement.setAttribute(AttributeType.TIME_MIN, min);
		replacement.setAttribute(AttributeType.TIME_MAX, max);
		replacement.setAttribute(AttributeType.TIME_AVG, avg);
		
		return (FlowNode)replacement;
	}

	/**
	 * {@inheritDoc}
	 * @throws ReplacementException 
	 */
	@Override
	public FlowNode replaceCycle(final PatternMatcherResult result) throws ReplacementException {
		//create new node and set and fetch default values
		IContainAttribute replacement = new CustomFunction(result.getEntry().getLabel()+"~"+result.getExit().getLabel());
		addMissingAttributeValues(replacement);
		double min = 0;
		double max = Double.POSITIVE_INFINITY;
		double avg = 0;
			
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = result.getOwner().getEpc();

		//follow main path and get values
		IContainAttribute mainPathNode = ((IContainAttribute)epc.getFirstDirectSuccessor(result.getEntry()));
		double minMainPath = (double) mainPathNode.getAttribute(AttributeType.TIME_MIN);
		double maxMainPath = (double) mainPathNode.getAttribute(AttributeType.TIME_MAX);
		double avgMainPath = (double) mainPathNode.getAttribute(AttributeType.TIME_AVG);
		
		//follow split's outgoing paths to find back path
		for(ControlFlow<FlowNode> edge : epc.getOutgoingEdges(result.getExit())) {
			
			if(epc.getFirstDirectSuccessor(edge.getTarget())==result.getEntry()) {
				//this is the back path
				IContainAttribute backPathNode = (IContainAttribute)edge.getTarget();
				IContainAttribute backPathEdge = (IContainAttribute) edge;
				double minBackPath = (double) backPathNode.getAttribute(AttributeType.TIME_MIN);
				double maxBackPath = (double) backPathNode.getAttribute(AttributeType.TIME_MAX);
				double avgBackPath = (double) backPathNode.getAttribute(AttributeType.TIME_AVG);
				
				min = minMainPath; //main path is always executed (at least once)
				
				//calculate as geometric series
				if(backPathEdge.hasAttribute(AttributeType.PROBABILITY)) {
					double probability = (double) backPathEdge.getAttribute(AttributeType.PROBABILITY);
					if(Math.max(maxMainPath, maxBackPath)==0) { max=0; } //cycle is timeless then
					else { max=Double.POSITIVE_INFINITY; } //infinite repeating causes infinite time
					//a hint to understand the calculation
					//sum[from i=0 to infinity](a*p^i) for 0 < p < 1, a=const
					//= a * (1-lim[i->infinity](p^i)) / (1-p)
					//= a * (1-0) / (1-p)
					//sum[from i=1 to infinity](a*p^i) = a * ((1/(1-p))-1) 
					avg = avgMainPath + (avgMainPath+avgBackPath) * ((1/(1-probability))-1);
				}
				//calculate using tuple
				else if(backPathEdge.hasAttribute(AttributeType.PROBABILITY_LIST)) {
//					System.out.println(backPathEdge.getAttribute(AttributeType.PROBABILITY_LIST));
					List<Double> probabilities = (List<Double>) backPathEdge.getAttribute(AttributeType.PROBABILITY_LIST);
					max = maxMainPath + (maxMainPath+maxBackPath)*probabilities.size();
					
					avg = avgMainPath;
					double fullPathProbability = 1.0;
					for(double probability : probabilities) { //probability is the chance to cycle back in the current split
						fullPathProbability *= probability; //this is the overall chance to cycle back in the n-th path
						avg += (avgMainPath+avgBackPath) * fullPathProbability;
					}
				}
				//cannot calculate without probabilities
				else {
					throw new ReplacementException(Message.get("verifier.probability_value_missing_on_cycle"), result.getEntry());
				}
				break; //only need the cyclic path, skip others
			}
			else { //on none cyclic paths perform some checks
				if(((IContainAttribute) edge).hasAttribute(AttributeType.PROBABILITY_LIST)) {
					throw new ReplacementException(Message.get("verifier.probability_list_on_wrong_edge"), result.getEntry());
				}
			}
		}
		
		//write values to the new element
		replacement.setAttribute(AttributeType.TIME_MIN, min);
		replacement.setAttribute(AttributeType.TIME_MAX, max);
		replacement.setAttribute(AttributeType.TIME_AVG, avg);
		
		return (FlowNode)replacement;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void addMissingAttributeValues(final IContainAttribute node) {
		Double time;
		if(node.hasAttribute(AttributeType.TIME)) {
			time = (Double) node.getAttribute(AttributeType.TIME);
		}
		else {
			time = 0.0;
			node.setAttribute(AttributeType.TIME, time);
		}
		//prepare min, max, avg
		node.setAttribute(AttributeType.TIME_MIN, time);
		node.setAttribute(AttributeType.TIME_MAX, time);
		node.setAttribute(AttributeType.TIME_AVG, time);
	}
	
}
