package de.fh_zwickau.bumb_l_b.qos_tool.processor.pattern;

import java.util.HashSet;
import java.util.Iterator;

import org.jbpt.pm.ControlFlow;
import org.jbpt.pm.FlowNode;
import org.jbpt.pm.NonFlowNode;
import org.jbpt.pm.epc.IEpc;

import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper.AttributeType;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomAndConnector;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomEvent;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomFunction;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomOrConnector;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomXorConnector;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.IContainAttribute;

/**
 * Base class for specialized pattern matcher. Pattern matcher are used to find regions by a pattern within an epc. 
 */
public abstract class PatternMatcher {
	protected IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc;
	protected HashSet<PatternMatcherResult> results;
	
	/**
	 * Creates a new pattern matcher.
	 * @param epc
	 */
	public PatternMatcher(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc) {
		setEpc(epc);
		results = new HashSet<PatternMatcherResult>();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void find() {
		results.clear(); //clear all results
		findEntryPoints(); //Find entry node where and regions could start
		siftResults(); //sift out the results
		//printResults(); //only for debugging
	}
	
	/**
	 * Find nodes where a matching region could possibly start. A result is generated for every possible entry point.
	 */
	protected abstract void findEntryPoints();
	
	/**
	 * Sift out non matching results.
	 */
	protected void siftResults() {
		Iterator<PatternMatcherResult> it = results.iterator();
		while(it.hasNext()) {
			PatternMatcherResult result = it.next();
			
			if(!isMatchingResult(result)) {
				it.remove();
			}
		}
	}
	
	/**
	 * Checks whether the given result matches the pattern of this matcher. In the result the affected nodes are updated as well as the exit
	 * and post exit nodes.
	 * @param result
	 * @return true: result matches, false: result doesn't match
	 */
	protected abstract boolean isMatchingResult(PatternMatcherResult result);
	
	/**
	 * Prints out all results
	 */
	public void printResults() {
		for(PatternMatcherResult r : results) {
			System.out.println(r);
		}
	}
	
	/**
	 * Sets the epc in which the pattern will be searched
	 * @param epc
	 */
	public void setEpc(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc) {
		this.epc = epc;
	}
	
	/**
	 * Returns the epc in which the pattern is searched
	 * @return
	 */
	public IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> getEpc() {
		return epc;
	}
	
	/**
	 * Returns a HashSet containing all results.
	 * @return
	 */
	public HashSet<PatternMatcherResult> getResults() {
		return new HashSet<PatternMatcherResult>(results);
	}

	/**
	 * Inner class to hold results
	 */
	public class PatternMatcherResult {
		protected FlowNode preEntry;
		protected FlowNode entry;
		protected FlowNode exit;
		protected FlowNode postExit;
		protected HashSet<FlowNode> nodes;
		protected PatternMatcher owner;
		
		/**
		 * Creates a new result
		 */
		public PatternMatcherResult(PatternMatcher matcher) {
			this(matcher, null, null);
		}
		
		/**
		 * Creates a new result
		 * @param preEntry after that node the matching region starts
		 * @param entry first node of the matching region
		 */
		public PatternMatcherResult(PatternMatcher matcher, FlowNode preEntry, FlowNode entry) {
			this(matcher, preEntry, entry, null, null);
		}
		
		/**
		 * Creates a new result
		 * @param owner the matcher that generated the result
		 * @param preEntry after that node the matching region starts
		 * @param entry first node of the matching region
		 * @param exit last node of the matching region
		 * @param postExit next node after the matching region
		 */
		public PatternMatcherResult(PatternMatcher owner, FlowNode preEntry, FlowNode entry, FlowNode exit, FlowNode postExit) {
			this.owner = owner;
			this.preEntry = preEntry;
			this.entry = entry;
			this.exit = exit;
			this.postExit = postExit;
			nodes = new HashSet<FlowNode>();
		}
		
		/**
		 * Returns the node after which the matching region starts
		 * @return
		 */
		public FlowNode getPreEntry() {
			return preEntry;
		}

		/**
		 * Returns the first node of the matching region
		 * @return
		 */
		public FlowNode getEntry() {
			return entry;
		}

		/**
		 * Returns the first edge within the match. That is important for fetching probabilities.
		 * @return
		 */
		public ControlFlow<FlowNode> getEntryEdge() {
			if(preEntry != null) {
				if(entry != null) {
					return owner.epc.getDirectedEdge(preEntry, entry);
				}
				else if(exit == null && postExit != null) { //because empty pattern do neither have entry nor exit
					return owner.epc.getDirectedEdge(preEntry, postExit);
				}
			}
			return null;
		}
		
		/**
		 * Returns the last node of the matching region
		 * @return
		 */
		public FlowNode getExit() {
			return exit;
		}

		/**
		 * Returns the next node after the matching region
		 * @return
		 */
		public FlowNode getPostExit() {
			return postExit;
		}

		/**
		 * Returns the matcher that generated the result
		 * @return
		 */
		public PatternMatcher getOwner() {
			return owner;
		}
		
		/**
		 * Returns all affected nodes i.e. nodes that are within the matching region.
		 * @return
		 */
		public HashSet<FlowNode> getNodes() {
			return new HashSet<FlowNode>(nodes);
		}
		
		@Override
		public String toString() {
			return String.format("(%s)-%s-...-%s-(%s) #%d", formatNodeInfo(preEntry), formatNodeInfo(entry), formatNodeInfo(exit), formatNodeInfo(postExit), nodes.size());
		}
		
		/**
		 * Helper for the toString method
		 * @param node
		 * @return
		 */
		private String formatNodeInfo(FlowNode node) {
			IContainAttribute attrNode = (IContainAttribute) node;
			if(node == null) { return "null"; }
			String info = node.getLabel();
			if("".equals(info)) {
				if(node instanceof CustomAndConnector) { info = "and"; }
				else if(node instanceof CustomEvent ) { info = "evt"; }
				else if(node instanceof CustomFunction ) { info = "fnc"; }
				else if(node instanceof CustomXorConnector ) { info = "xor"; }
				else if(node instanceof CustomOrConnector ) { info = "or"; }
				info+=(String) attrNode.getAttribute(AttributeType.BFLOW_ID);
			}
			return info;
		}
	}
}
