package de.fh_zwickau.bumb_l_b.qos_tool.verifier;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import org.jbpt.pm.ControlFlow;
import org.jbpt.pm.FlowNode;
import org.jbpt.pm.NonFlowNode;
import org.jbpt.pm.epc.IEpc;

import de.fh_zwickau.bumb_l_b.qos_tool.Message;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper.AttributeType;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomControlFlow;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomAndConnector;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomFunction;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomOrConnector;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.CustomXorConnector;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.IContainAttribute;
import de.fh_zwickau.bumb_l_b.qos_tool.printer.IPrinter;

/**
 * Checking attributes of nodes of epc models. 
 * 
 * @author Marcel Liebl
 *
 */
public class EpcVerifier implements IEpcVerifier
{
	private Set<AttributeType> calculatableAttributeTypes;
	private IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc;
	private IPrinter printer;
	
	private HashMap<Integer, Double> tupelValues = new HashMap<Integer,Double>();
	
	private boolean hasAnyCostAttributes = false;
	private boolean hasAnyTimeAttributes = false;
	private boolean hasAnyReliabilityAttributes = false;
	private boolean hasAnyProbabilityAttributes = false;
	private boolean hasAnyNOutOfMAttributes = false;
	private boolean hasAnyProbabilityListAttributes = false;
	
	private boolean epcRejected = false;
	
	private boolean costCanBeCalc = true;
	private boolean timeCanBeCalc = true;
	private boolean reliabilityCanBeCalc = true;
	
	/**
	 * Constructor - gets instance of printer.
	 * 
	 *  @param IPrinter printer
	 */
	public EpcVerifier(IPrinter p)
	{
		printer = p;
	}

	
	/**
	 * Checking nodes of the given epc. Prints messages if attributes are null or out of bounds.
	 * 
	 * @param IEpc e
	 */
	@Override
	public boolean verify(IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc)
	{
		calculatableAttributeTypes = new HashSet<AttributeType>();
		this.epc = epc;
		
//Checking functions
		
		//Check if functions got any attributes
		for(FlowNode fu : epc.getFunctions())
		{
			CustomFunction Cfu = (CustomFunction) fu;
			hasAnyAttributes(Cfu);
		}
		
		//Check if attributes of functions are in bounds
		for(FlowNode fu : epc.getFunctions())
		{
			CustomFunction Cfu = (CustomFunction) fu;
		
			checkAttributes(Cfu, Cfu.getName(), Cfu.getAttribute(AttributeType.BFLOW_ID).toString());

		}
		
//Checking Gateways
		
		
		//Checking if gateways got attributes (if and connector got n-out-of-m / if xor got probability or probability list
		for(FlowNode n : epc.getGateways())
		{
			if(epc.getGateways().isEmpty() != true)
			{
				if(isXor(n, epc))
				{
					CustomXorConnector Cxor = (CustomXorConnector) n;
		
					xorEdgesHaveAttributes(Cxor);
				}
								
				if(isAnd(n))
				{
					CustomAndConnector Cand = (CustomAndConnector) n;
					
					andConnectorGotAttributes(Cand, AttributeType.N_OUT_OF_M);
				}	
			}
				
		}
		
		
		//Checking attributes of the gateways
		for(FlowNode n : epc.getGateways())
		{	
			if(epc.getGateways().isEmpty() != true)
			{
				if(isXor(n, epc))
				{
					CustomXorConnector Cxor = (CustomXorConnector) n;
					
					checkXorAttributes(Cxor);
				}
				
				if(isAnd(n))
				{
					CustomAndConnector Cand = (CustomAndConnector) n;
					
					checkAndAttributes(Cand);
				}
				
				if(isOrGateway(n))
				{
					CustomOrConnector Cor = (CustomOrConnector) n;
					
					printer.printError(Message.get("verifier.model_contains_or"),Cor.getId());
					
					epcRejected = true;
				}
			}
		}
		
//Can attributes be calculated
		
		if(hasAnyCostAttributes)
		{
			if(costCanBeCalc)
			{
				calculatableAttributeTypes.add(AttributeType.COST);
			}
			
			else
			{
				epcRejected = true;
			}
		}
		
		if(hasAnyTimeAttributes)
		{
			if(timeCanBeCalc)
			{
				calculatableAttributeTypes.add(AttributeType.TIME);
			}
			
			else
			{
				epcRejected = true;
			}
		}
		
		if(hasAnyReliabilityAttributes)
		{
			if(reliabilityCanBeCalc)
			{
				calculatableAttributeTypes.add(AttributeType.RELIABILITY);
			}
			
			else
			{
				epcRejected = true;
			}
		}
		
		
	//accept or decline of epc	
		if(epcRejected)
		{
			return false;
		}
		
		else
		{
			return true;
		}
		
	}
	
	/**
	 * Checking if any functions got any attributes.
	 * 
	 * @param IContainAttribute i
	 * 
	 */
	public void hasAnyAttributes(IContainAttribute i)
	{
	//has cost attribute
		if(checkValue(i,AttributeType.COST) != null)
		{
			hasAnyCostAttributes = true;
		}
	
	//has time attribute
		if(checkValue(i,AttributeType.TIME) != null)
		{
			hasAnyTimeAttributes = true;
		}
	
	//has reliability attribute
		if(checkValue(i,AttributeType.RELIABILITY) != null)
		{
			hasAnyReliabilityAttributes = true;
		}
		
	}
	
	/**
	 * Checking if xor edges got attributes.
	 * 
	 * @param CustomXorConnector xor
	 */
	public void xorEdgesHaveAttributes(CustomXorConnector xor)
	{
		for(ControlFlow<FlowNode> c  : epc.getOutgoingEdges(xor))
		{
			CustomControlFlow cust = (CustomControlFlow) c;
			
			String probList =  checkProbabilityListGotAttribute(cust, AttributeType.PROBABILITY_LIST);
			
			//xor edge got probability list -> true or false
			if(probList != null)
			{
				hasAnyProbabilityListAttributes = true;
			}
				
			//xor edge got probability 
			Double attVp = checkValue(cust,AttributeType.PROBABILITY);
				
			if(attVp != null)
			{
				hasAnyProbabilityAttributes = true;
			}
				
			
		}
	}
	
	/**
	 * Checking if and connector got attributes.
	 * 
	 */
	public void andConnectorGotAttributes(CustomAndConnector and, AttributeType type)
	{
		if(checkAndConnectorValue(and,type) != null)
		{
			hasAnyNOutOfMAttributes = true;
		}
	}
	
	/**
	 * Checking attributes of events and functions.
	 * 
	 * @param IContainAttribut i
	 * @param String name
	 * 
	 */
	public void checkAttributes(IContainAttribute i, String name, String id)
	{
		
//Checking cost
		if(hasAnyCostAttributes)
		{			
			Double attVc = checkValue(i,AttributeType.COST);
		

			if(attVc != null)
			{	
				if(attVc.isNaN())
				{
					printer.printError(String.format(Message.get("verifier.attribute_value_not_numeric"), "cost"), id);
					
					costCanBeCalc = false;
					epcRejected = true;
				}
				
				if(attVc < 0)
				{	
					printer.printError(Message.get("verifier.cost_lower_than_0"), id);
					
				
					costCanBeCalc = false;
					
					epcRejected = true;
				}			
			
			}
			
			else
			{			
				printer.printError(String.format(Message.get("verifier.incomplete_qos_attribute"), "cost", "cost"),id);
				
				costCanBeCalc = false;
				
			}		
			
		}
		
//Checking time
		if(hasAnyTimeAttributes)
		{
			Double attVt = checkValue(i,AttributeType.TIME);
		
			if(attVt != null)
			{	
				if(attVt.isNaN())
				{
					printer.printError(String.format(Message.get("verifier.attribute_value_not_numeric"), "time"), id);
					
					timeCanBeCalc = false;
					epcRejected = true;
				}
				
				if(attVt < 0)
				{
					printer.printError(Message.get("verifier.time_lower_than_0"), id);
				
					timeCanBeCalc = false;
				
					epcRejected = true;
				}
			}
		
			else
			{
					printer.printError(String.format(Message.get("verifier.incomplete_qos_attribute"), "time", "time"),id);
				
					timeCanBeCalc = false;
			}
			
		}
		
//Checking reliability
		if(hasAnyReliabilityAttributes)
		{
			Double attVr = checkValue(i,AttributeType.RELIABILITY);
		
			if(attVr != null)
			{		
				if(attVr.isNaN())
				{
					printer.printError(String.format(Message.get("verifier.attribute_value_not_numeric"), "reliability"), id);
					
					reliabilityCanBeCalc = false;
					epcRejected = true;
				}
				
				if(attVr < 0 || attVr == 0  || attVr > 1)
				{		
					printer.printError(Message.get("verifier.reliability_limits_exceeded"), id);
				
					reliabilityCanBeCalc = false;
				
					epcRejected = true;
				}
			}
		
			else
			{			
				printer.printError(String.format(Message.get("verifier.incomplete_qos_attribute"), "reliability", "reliability"),id);
				
				reliabilityCanBeCalc = false;
				
			}		
		}
	}
	
	/**
	 * Checking if FlowNode is Xor-connector
	 * 
	 * @param FlowNode n
	 */
	public boolean isXor(FlowNode n, IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc)
	{
		
		try
		{
			CustomXorConnector Cxor = (CustomXorConnector) n;
			
			return true;
			
		}
		catch(Exception e)
		{
			return false;
		}	
	}
	
	/**
	 * Checking if FlowNode is an and-connector.
	 * 
	 * @param FlowNode n
	 */
	public boolean isAnd(FlowNode n)
	{
		try
		{
			CustomAndConnector Cand = (CustomAndConnector) n;
			
			return true;
		}
		catch(Exception e)
		{
			return false;
		}	
	}
	
	/**
	 * Checking if FlowNode is an or-connector.
	 * 
	 * @param FlowNode n
	 */
	public boolean isOrGateway(FlowNode n)
	{
		try
		{
			CustomOrConnector Cor = (CustomOrConnector) n;
			
			return true;
			
		}
		catch(Exception e)
		{
			return false;
		}	
	}

	/**
	 * Splits the probability list and check the value of the attributes.
	 * 
	 * @param c
	 * @param s
	 */
	public void checkProbabilityListValue(Integer outgoingEdges, String s)
	{
		String values = s;
		tupelValues = new HashMap<Integer, Double>();
		Integer key = 0;
		
		values = values.replace("[", "");
		values = values.replace("]", "");
		values = values.replace(",", " ");
		
		String[] list = values.split(" ");	

		if(values.isEmpty())
		{
			printer.printError(Message.get("verifier.probability_list_is_empty"));
			
			epcRejected = true;
		}
		
		for(String i : list)
		{
			key++;
			
			checkProbabilityListAttributes(i, outgoingEdges, key);
		}
		
		//we get the values of number 1,3,5 because if adds always a null at even numbers
		Double tupelValue1 = tupelValues.get(1);
		Double tupelValue3 = tupelValues.get(3);
		Double tupelValue5 = tupelValues.get(5);
		
		if(tupelValue1 < tupelValue3 || tupelValue3 < tupelValue5 || tupelValue1 < tupelValue5)
		{
			printer.printWarning( Message.get("verifier.probability_list_not_descending"));
		}
	
	}
	
	/**
	 * Checking prob list attributes
	 * @param s
	 */
	public void checkProbabilityListAttributes(String s, Integer outgoingEdges, Integer key)
	{
		try
		{
			Double value = Double.parseDouble(s);
			tupelValues.put(key, value);
			
			if(value.isNaN())
			{
				printer.printError(Message.get("verifier.probability_list_contains_not_numeric"));
				
				epcRejected = true;
				
			}
				
			if(value <= 0 || value >= 1)
			{
				printer.printError(Message.get("verifier.probability_list_limits_exceeded"));
				
				epcRejected = true;
			}
			
			if(outgoingEdges > 2 || outgoingEdges < 2)
			{
				
				printer.printError(Message.get("verifier.xor_cycle_wrong_number_of_edges"));
				
				epcRejected = true;
		
			}

			
		}
		
		catch(NumberFormatException e)
		{		
//			epcRejected = true;
//			
//			probListCanBeCalc = false;
		}
		
	}
	
	/**
	 * Checking probability attribute at xor-connector.
	 * 
	 * @param CustomXorConnector i
	 * 
	 */
	public void checkXorAttributes(CustomXorConnector i)
	{
		double value = 0;
		boolean gotProbList = false;
		Integer outgoingEdges = 0;
		Integer countingLists = 0;
	
	//Checking if any edges of this xor connector got a probability list
		for(ControlFlow<FlowNode> c  : epc.getOutgoingEdges(i))
		{
			CustomControlFlow cust = (CustomControlFlow) c;
			
			if(hasAnyProbabilityListAttributes)
			{
				if(cust.getAttribute(AttributeType.PROBABILITY_LIST) != null)
				{
					// if xor connector got a list at any outgoing edges
					gotProbList = true;
					
					//counting lists at all edges
					countingLists ++;
				}
			}
		}
		
	//Checking attributes of edges
		for(ControlFlow<FlowNode> c  : epc.getOutgoingEdges(i))
		{
			CustomControlFlow cust = (CustomControlFlow) c;
			outgoingEdges = countOutgoingEdges(i);
			
			if(hasAnyProbabilityAttributes)
			{
					//try to cast probability in double
					Double attVp = checkValue(cust,AttributeType.PROBABILITY);
					
					if(attVp != null)
					{	
						
						//Check if xor connector got a probList at other outgoing edge
						if(gotProbList)
						{
							printer.printError(Message.get("verifier.probability_single_and_list_mixed"));
							
							epcRejected = true;
						}
						
						else
						{
							// if probability value is not a number
							if(attVp.isNaN())
							{
								printer.printError(Message.get("verifier.probability_not_numeric_or_tuple"),getXorId(i));
							
								epcRejected = true;
							}
						
							// value = sum value of all edges of the xor connector
							value = value + attVp;
						
							//checking probability value
							if(attVp < 0 || attVp > 1 || attVp == 0 || attVp == 1 )
							{	
								printer.printError(Message.get("verifier.probability_limits_exceeded"),getXorId(i));
							
								epcRejected = true;
							}
						}
						
						
					}
					
					//If xor got only one outgoing edge, the sum value is set to one
					if(outgoingEdges == 1)
					{
						value = 1;
					}
			
				}
		
			// if controlflow got no probability attribute..
			else
			{
				//check if he got probList attributes...
				if(hasAnyProbabilityListAttributes)
				{
					//casting probList into string
					try
					{
						String list = cust.getAttribute(AttributeType.PROBABILITY_LIST).toString();
						
						checkProbabilityListValue(outgoingEdges, list);
					}
					
					catch(NullPointerException e)
					{
						
					}	

				}
			
				//otherwise probability values are missing
				else
				{
					if(epc.getIncomingEdges(i).size() == 1)
					{
						printer.printError(Message.get("verifier.probability_value_missing"), getXorId(i));
						epcRejected = true;
					}
				}
			}
		}

		if(hasAnyProbabilityAttributes)
		{
			
			//if edge got invalid sum value, and the xor connector got no list on other edges
				if(value > 1 || value < 1 && gotProbList == false)
				{
					printer.printError(Message.get("verifier.probability_single_sum_fault"),getXorId(i));
					
					epcRejected = true;
				}

		}
		
		//xor connector got more than 2 lists 
		if(countingLists == 2)
		{
			printer.printError(Message.get("verifier.probability_list_more_than_one_list"));
			
			epcRejected = true;
		}

	}
	
	/**
	 * Counts the outgoing edges of a node. 
	 * 
	 * @param IContainAttribute att
	 * @return Integer i
	 */
	public Integer countOutgoingEdges(IContainAttribute att)
	{
		int i = 0;
		for(ControlFlow<FlowNode> c : epc.getOutgoingEdges((FlowNode) att))
		{	
			i++;
		}
		
		return i;
		
	}
	
	/**
	 * Counting incomming edges of a node.
	 * 
	 * @param IContainAttribute att
	 */
	public Integer countIncommingEdges(IContainAttribute att)
	{
		int i = 0;
		
		for(ControlFlow<FlowNode> c  : epc.getIncomingEdges((FlowNode) att))
		{	
			i++;
		}
		
		return i;
	}
	
	
	/**
	 * Checking attributes at and-connector.
	 * 
	 * @param CustomAndConnector i
	 * 
	 */
	public void checkAndAttributes(CustomAndConnector i)
	{	
		if(hasAnyNOutOfMAttributes)
		{
			//getting value of n-out-of-m attribute
			Integer attVn = checkAndConnectorValue(i,AttributeType.N_OUT_OF_M);
			
			Integer edges = countIncommingEdges(i);
			Integer edgesOutgoing = countOutgoingEdges(i);
			
			
			if(attVn != null)
			{	
					if(attVn < 1)
					{
						printer.printError(Message.get("verifier.n_lower_than_1"),getAndId(i));
					
						epcRejected = true;
					}
				
					if(attVn > edges)
					{
						printer.printError(Message.get("verifier.n_greater_than_m"),getAndId(i));
					
						epcRejected = true;
					}
				
					if(edgesOutgoing > 1)
					{
						printer.printError(Message.get("verifier.n_on_split"), getAndId(i));
					
						epcRejected = true;
					}
			}
		}
		
	}
	
	/**
	 * Gets name of an attribute and parse the value from object to double.
	 * 
	 * @param	IContainAttribute i, String s
	 * 
	 */
	public Double checkValue(IContainAttribute i , AttributeType type)
	{
		
		if(AttributeHelper.isAttributeOfType(type, i.getAttribute(type)))
		{
			
				return Double.parseDouble(i.getAttribute(type).toString());
	
		}
		
		else
		{	
			return null;
		}
		
	}
	
	/**
	 * Checking if probability list got any attributes.
	 * 
	 * @param i
	 * @param type
	 */
	public String checkProbabilityListGotAttribute(IContainAttribute i, AttributeType type)
	{
		if(AttributeHelper.isAttributeOfType(type, i.getAttribute(type)))
		{
				return i.getAttribute(type).toString();
	
		}
		
		else
		{	
			return null;
		}
		
	}
	
	/**
	 * Checking value of and connector.
	 * 
	 * @param	IContainAttribute i, AttributeType type
	 */
	public Integer checkAndConnectorValue(IContainAttribute i , AttributeType type)
	{
		
		if(AttributeHelper.isAttributeOfType(type, i.getAttribute(type)))
		{
			Integer integ = Integer.parseInt(i.getAttribute(type).toString());
			
			//if integer is NaN, the parser set his value to min, so it can be checked here
			if(Integer.MIN_VALUE != integ)
			{
				return integ;
			}
			
			else
			{
				printer.printError(Message.get("verifier.n_not_integer"));
				
				epcRejected = true;
				
				return null;
			}
			
		}

		else
		{
			return null;
		}
	}
	
	
	/**
	 * Return the id of a Xor connector.
	 * 
	 * @param CustomXorConnector i
	 *  
	 */
	public String getXorId(CustomXorConnector i)
	{
		return i.getAttribute(AttributeType.BFLOW_ID).toString();
	}
	
	/**
	 * Return the id of an and connector.
	 * 
	 * @param CustomAndConnector a
	 */
	public String getAndId(CustomAndConnector a)
	{
		return a.getAttribute(AttributeType.BFLOW_ID).toString();
	}

	/**
	 * Returns set of attribute types.
	 * 
	 * @return Set<AttributeType>
	 */
	@Override
	public Set<AttributeType> getProcessableAttributeTypes() 
	{
		return calculatableAttributeTypes;
	}
}
