package de.fh_zwickau.bumb_l_b.qos_tool.processor;

import static org.junit.Assert.*;

import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.jbpt.pm.ControlFlow;
import org.jbpt.pm.FlowNode;
import org.jbpt.pm.NonFlowNode;
import org.jbpt.pm.epc.IEpc;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.Timeout;
import de.fh_zwickau.bumb_l_b.qos_tool.Message;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper.AttributeType;
import de.fh_zwickau.bumb_l_b.qos_tool.loader.EpcLoaderException;
import de.fh_zwickau.bumb_l_b.qos_tool.loader.EpmlEpcLoader;
import de.fh_zwickau.bumb_l_b.qos_tool.printer.AbstractPrinter;
import de.fh_zwickau.bumb_l_b.qos_tool.printer.AbstractPrinter.MsgType;

public class QosProcessorTest {
	private static final TestPrinter printer = new TestPrinter();
	private static EpmlEpcLoader loader;
	private QoSProcessor processor;
	private IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc;
	private static final double ATTRIBUTE_DELTA = 1e-7;

    @Rule
    public Timeout globalTimeout = new Timeout(500); // max ms per method
//    public Timeout globalTimeout = new Timeout(100); // max ms per method
	
	@BeforeClass
	public static void setUpBeforeClass() throws Exception {
		loader = new EpmlEpcLoader();
	}

	@Before
	public void setUp() throws Exception {
		printer.reset();
		processor = new QoSProcessor(printer);
	}

	private void load(String name) {
		try {
			epc = loader.loadFile("testContent/processor/"+name+".epml");
		} catch (EpcLoaderException e) {
			fail("Load file failed");
		}
	}
	
	private static class TestPrinter extends AbstractPrinter {
		ArrayList<String> out = new ArrayList<String>();
		
		public TestPrinter() {}
		
		public void reset() {
			out.clear();
		}
		
		@Override
		protected void printMessage(MsgType type, String message, String elementId) {
			out.add(formatMessage(type, message, elementId));
		}
		
		public static String formatMessage(MsgType type, String message, String elementId) {
			return String.format("%s: %s", type, message);
		}
		
		public String getMessage(int index) {
			if(out.isEmpty()) { return ""; }
			return out.get(index);
		}
		
		public void printOut() {
			for(String msg : printer.out) {
				System.out.println(msg);
			}
		}

		public void testNoErrorOrWarning() {
			for(String msg : out) {
				if(msg.startsWith(MsgType.WARNING.toString()) || msg.startsWith(MsgType.ERROR.toString())) {
					fail("Found Message "+msg);
				}
			}
		}
		
		/**
		 * Durchsucht den Pufferinhalt nach einem Muster und versucht das erste Teilergebnis im Muster als Double zu parsen.
		 * @param pattern Regex-Suchmuster
		 * @return Gefundener Wert
		 */
		private Double parseDoubleByPattern(String pattern) {
			Pattern p = Pattern.compile(pattern);
			Matcher m;
			boolean found = false;
			
			Double value = null;
			
			for(String msg : out) {
				m = p.matcher(msg);
				try {
					if(m.find(1)) {
						if(found) { fail("More than one result found for the current pattern!"); }
						value = new Double(m.group(1));
						found = true;
					}
				} catch (Exception e) {
					fail("Result pattern not found");
				}
			}
			
			if(!found) {
				fail("Result pattern not found");				
			}
			
			return value;
		}
		
		public Double getMinTime() {
			return parseDoubleByPattern(String.format(Message.get("processor.result"), "TIME", "min", "(\\d+\\.\\d+)"));
		}
		public Double getMaxTime() {
			return parseDoubleByPattern(String.format(Message.get("processor.result"), "TIME", "max", "(\\d+\\.\\d+)"));
		}
		public Double getAvgTime() {
			return parseDoubleByPattern(String.format(Message.get("processor.result"), "TIME", "avg", "(\\d+\\.\\d+)"));
		}
		public Double getMinCost() {
			return parseDoubleByPattern(String.format(Message.get("processor.result"), "COST", "min", "(\\d+\\.\\d+)"));
		}
		public Double getMaxCost() {
			return parseDoubleByPattern(String.format(Message.get("processor.result"), "COST", "max", "(\\d+\\.\\d+)"));
		}
		public Double getAvgCost() {
			return parseDoubleByPattern(String.format(Message.get("processor.result"), "COST", "avg", "(\\d+\\.\\d+)"));
		}
		public Double getMinReliability() {
			return parseDoubleByPattern(String.format(Message.get("processor.result"), "RELIABILITY", "min", "(\\d+\\.\\d+)"));
		}
		public Double getMaxReliability() {
			return parseDoubleByPattern(String.format(Message.get("processor.result"), "RELIABILITY", "max", "(\\d+\\.\\d+)"));
		}
		public Double getAvgReliability() {
			return parseDoubleByPattern(String.format(Message.get("processor.result"), "RELIABILITY", "avg", "(\\d+\\.\\d+)"));
		}		
		
	};
	
	@Test
	public void test_minimal() throws EpcLoaderException {
		load("minimal");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(10, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(10, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(10, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_empty() throws EpcLoaderException {
		load("empty");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("processor.calculation_failed"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_linear1f() throws EpcLoaderException {
		load("linear1f");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(10, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(10, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(10, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}

	@Test
	public void test_linear3f() throws EpcLoaderException {
		load("linear3f");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(33, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(33, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(33, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.504, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.504, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.504, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3.3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(3.3, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(3.3, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_1path_1empty() throws EpcLoaderException {
		load("and_1path_1empty");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(10, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(10, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(10, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_1path() throws EpcLoaderException {
		load("and_1path");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(10, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(10, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(10, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_2path_1empty_1outof2() throws EpcLoaderException {
		load("and_2path_1empty_1outof2");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(10, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(10, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(10, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.95, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(0, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(0, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(0, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_2path_1empty() throws EpcLoaderException {
		load("and_2path_1empty");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(10, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(10, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(10, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_2path_1nestedand() throws EpcLoaderException {
		load("and_2path_1nestedand");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(60, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(60, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(60, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.504, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.504, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.504, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_2path_1nestedlinear() throws EpcLoaderException {
		load("and_2path_1nestedlinear");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(60, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(60, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(60, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.504, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.504, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.504, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(5, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(5, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(5, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_2path_2nestedand() throws EpcLoaderException {
		load("and_2path_2nestedand");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(100, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(100, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(100, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.3024, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.3024, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.3024, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(4, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(4, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(4, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_2path_2nestedandlinear() throws EpcLoaderException {
		load("and_2path_2nestedandlinear");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(100, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(100, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(100, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.3024, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.3024, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.3024, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(7, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(7, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(7, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_2path() throws EpcLoaderException {
		load("and_2path");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(30, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(30, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.72, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.72, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.72, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(2, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(2, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(2, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_5path_1empty() throws EpcLoaderException {
		load("and_5path_1empty");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(100, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(100, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(100, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.504, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.504, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.504, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(4, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(4, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(4, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_5path_1outof5() throws EpcLoaderException {
		load("and_5path_1outof5");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(150, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(150, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(150, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.6, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_5path_2empty() throws EpcLoaderException {
		load("and_5path_2empty");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(60, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(60, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(60, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.72, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.72, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.72, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_5path_3outof5() throws EpcLoaderException {
		load("and_5path_3outof5");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(150, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(150, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(150, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.336, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.72, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.5, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_5path_5outof5() throws EpcLoaderException {
		load("and_5path_5outof5");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(150, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(150, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(150, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.3024, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.3024, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.3024, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(5, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(5, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(5, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_and_5path() throws EpcLoaderException {
		load("and_5path");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(150, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(150, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(150, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.3024, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.3024, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.3024, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(5, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(5, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(5, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_cycle_single_backempty() throws EpcLoaderException {
		load("cycle_single_backempty");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(35, printer.getAvgCost(), ATTRIBUTE_DELTA);
//		assertEquals(0, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.72, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.6857, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(3.5, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_cycle_single_mainempty() throws EpcLoaderException {
		load("cycle_single_mainempty");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(10, printer.getMinCost(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(15, printer.getAvgCost(), ATTRIBUTE_DELTA);
//		assertEquals(0, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8571, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMinTime(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(1.5, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_cycle_single() throws EpcLoaderException {
		load("cycle_single");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(42.5, printer.getAvgCost(), ATTRIBUTE_DELTA);
//		assertEquals(0, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8372, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8080, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(4.25, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test(timeout=10000)
	public void test_cycle_tuple1() throws EpcLoaderException {
		load("cycle_tuple1");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(80, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(45, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.7163, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8372, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8009, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(8, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(4.5, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test(timeout=10000)
	public void test_cycle_tuple10() throws EpcLoaderException {
		load("cycle_tuple10");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(530, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(48.38, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.176, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8372, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.7941, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(53, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(4.838, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test(timeout=10000)
	public void test_cycle_tuple3_backempty() throws EpcLoaderException {
		load("cycle_tuple3_backempty");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(90, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(37.32, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.3686, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.72, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.6693, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(9, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(3.732, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_cycle_tuple3_mainempty() throws EpcLoaderException {
		load("cycle_tuple3_mainempty");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(10, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(70, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(17.32, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.4608, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8367, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(7, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(1.732, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_cycle_tuple3() throws EpcLoaderException {
		load("cycle_tuple3");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(180, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(48.3, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.5244, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8372, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.7942, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(18, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(4.83, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_xor_2path_1empty() throws EpcLoaderException {
		load("xor_2path_1empty");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(10, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(30, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(24, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.72, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.774, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(1, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(2.4, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_xor_2path_1nestedand() throws EpcLoaderException {
		load("xor_2path_1nestedand");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(130, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(100, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.7557, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8372, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.7802, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(9, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(7.2, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_xor_2path_1nestedand2() throws EpcLoaderException {
		load("xor_2path_1nestedand2");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(190, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(142, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.7255, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8372, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.7590, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(15, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(11.4, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_xor_2path_1nestedand3() throws EpcLoaderException {
		load("xor_2path_1nestedand3");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(100, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(79, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.8126, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8372, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.82, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(6, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(5.1, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_xor_2path_1nestedcycle() throws EpcLoaderException {
		load("xor_2path_1nestedcycle");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(92, printer.getAvgCost(), ATTRIBUTE_DELTA);
//		assertEquals(0, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8372, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.7836, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(9.2, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_xor_2path_1nestedcycle2() throws EpcLoaderException {
		load("xor_2path_1nestedcycle2");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(134, printer.getAvgCost(), ATTRIBUTE_DELTA);
//		assertEquals(0, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8372, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.7623, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(13.4, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_xor_2path_1nestedcycle3() throws EpcLoaderException {
		load("xor_2path_1nestedcycle3");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(71, printer.getAvgCost(), ATTRIBUTE_DELTA);
//		assertEquals(0, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8554, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8237, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(7.1, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_xor_2path_1nestedlinear() throws EpcLoaderException {
		load("xor_2path_1nestedlinear");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(80, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(65, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.378, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.72, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.4806, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(8, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(6.5, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_xor_2path_1nestedxor() throws EpcLoaderException {
		load("xor_2path_1nestedxor");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(90, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(68.5, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.7955, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8372, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.811, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(9, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(6.85, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_xor_2path_1nestedxor2() throws EpcLoaderException {
		load("xor_2path_1nestedxor2");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(150, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(110.5, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.7637, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8372, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.7886, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(15, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(11.05, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_xor_2path_1nestedxor3() throws EpcLoaderException {
		load("xor_2path_1nestedxor3");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(30, printer.getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(60, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(47.5, printer.getAvgCost(), ATTRIBUTE_DELTA);
		assertEquals(0.8372, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8645, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8531, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(3, printer.getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(6, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(4.75, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_xor_2path_3nestedcycle_weird() throws EpcLoaderException {
		load("xor_2path_3nestedcycle_weird");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		printer.testNoErrorOrWarning();
		assertEquals(110, printer.getMinCost(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(215.8, printer.getAvgCost(), ATTRIBUTE_DELTA);
//		assertEquals(0, printer.getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8205, printer.getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.7616, printer.getAvgReliability(), ATTRIBUTE_DELTA);
		assertEquals(11, printer.getMinTime(), ATTRIBUTE_DELTA);
//		assertEquals(Double.POSITIVE_INFINITY, printer.getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(21.58, printer.getAvgTime(), ATTRIBUTE_DELTA);
	}
	
	@Test
	public void test_probability_cycle_num_edges_gt_2() throws EpcLoaderException {
		load("probability_cycle_num_edges_gt_2");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("processor.calculation_failed"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_num_cycles_gt_1() throws EpcLoaderException {
		load("probability_num_cycles_gt_1");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("processor.calculation_failed"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_list_on_wrong_edge() throws EpcLoaderException {
		load("probability_list_on_wrong_edge");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		assertTrue(printer.getMessage(0).equals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_list_on_wrong_edge"), null)) || printer.getMessage(0).equals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_value_missing_on_cycle"), null)));
	}
	
	@Test
	public void test_probability_list_without_cycle() throws EpcLoaderException {
		load("probability_list_without_cycle");
		processor.process(epc, AttributeType.COST);
		processor.process(epc, AttributeType.RELIABILITY);
		processor.process(epc, AttributeType.TIME);
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_value_missing"), null), printer.getMessage(0));
	}
	
	
	
	
	
}

