package de.fh_zwickau.bumb_l_b.qos_tool.processor;

import static org.junit.Assert.*;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.jbpt.pm.ControlFlow;
import org.jbpt.pm.FlowNode;
import org.jbpt.pm.NonFlowNode;
import org.jbpt.pm.epc.IEpc;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper.AttributeType;
import de.fh_zwickau.bumb_l_b.qos_tool.loader.EpcLoaderException;
import de.fh_zwickau.bumb_l_b.qos_tool.loader.EpmlEpcLoader;
import de.fh_zwickau.bumb_l_b.qos_tool.printer.DefaultPrinter;

public class QosProcessorTestVarious {
	private static final double ATTRIBUTE_DELTA = 1e-7;
	
	private static final DefaultPrinter printer = DefaultPrinter.getInstance();
	private static final ByteArrayOutputStream bos = new ByteArrayOutputStream();
	private static final PrintStream ps = new PrintStream(bos);
	private static EpmlEpcLoader loader;
	private QoSProcessor processor;

	@BeforeClass
	public static void setUpBeforeClass() throws Exception {
		printer.setOut(ps);
		loader = new EpmlEpcLoader();
	}

	@Before
	public void setUp() throws Exception {
		bos.reset();
		processor = new QoSProcessor(printer);		
	}
	
	/**
	 * Durchsucht den Pufferinhalt nach einem Muster udn versucht das erste Teilergebnis im Muster als Double zu parsen.
	 * @param pattern Regex-Suchmuster
	 * @return Gefundener Wert
	 */
	private Double parseDoubleByPattern(String pattern) {
		Pattern p = Pattern.compile(pattern);
		Matcher m = p.matcher(bos.toString());
		Double value = null;
		try {
			if(m.find(1)) {
				value = new Double(m.group(1));
			}
		} catch (Exception e) {
			fail("Pattern not found");
		}
		return value;
	}
	
	private Double getMinTime() {
		return parseDoubleByPattern("TIME min: (\\d+\\.\\d+)");
	}
	private Double getMaxTime() {
		return parseDoubleByPattern("TIME max: (\\d+\\.\\d+)");
	}
	private Double getAvgTime() {
		return parseDoubleByPattern("TIME avg: (\\d+\\.\\d+)");
	}
	private Double getMinCost() {
		return parseDoubleByPattern("COST min: (\\d+\\.\\d+)");
	}
	private Double getMaxCost() {
		return parseDoubleByPattern("COST max: (\\d+\\.\\d+)");
	}
	private Double getAvgCost() {
		return parseDoubleByPattern("COST avg: (\\d+\\.\\d+)");
	}
	private Double getMinReliability() {
		return parseDoubleByPattern("RELIABILITY min: (\\d+\\.\\d+)");
	}
	private Double getMaxReliability() {
		return parseDoubleByPattern("RELIABILITY max: (\\d+\\.\\d+)");
	}
	private Double getAvgReliability() {
		return parseDoubleByPattern("RELIABILITY avg: (\\d+\\.\\d+)");
	}

	
	//detaillierte Tests für verschiedene Modelle (zu speichern in testContent)
	@Test
	public void test_process_test_tcr_and_small_2() throws EpcLoaderException {
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_tcr_and_small_2.epml");
		assertNotNull(epc);
		processor.process(epc, AttributeType.TIME);
		assertEquals(7, getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(7, getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(7, getAvgTime(), ATTRIBUTE_DELTA);
		bos.reset();
		processor.process(epc, AttributeType.COST);
		assertEquals(130, getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(130, getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(130, getAvgCost(), ATTRIBUTE_DELTA);
		bos.reset();
		processor.process(epc, AttributeType.RELIABILITY);
		assertEquals(0.7695, getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.7695, getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.7695, getAvgReliability(), ATTRIBUTE_DELTA);
		bos.reset();
	}
	
	@Test
	public void test_process_time_test_tcrn_and_small_1() throws EpcLoaderException {
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_tcrn_and_small_1.epml");
		assertNotNull(epc);
		processor.process(epc, AttributeType.TIME);
		assertEquals(20, getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(20, getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(20, getAvgTime(), ATTRIBUTE_DELTA);
		bos.reset();
		processor.process(epc, AttributeType.COST);
		assertEquals(300, getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(300, getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(300, getAvgCost(), ATTRIBUTE_DELTA);
		bos.reset();
		processor.process(epc, AttributeType.RELIABILITY);
		assertEquals(0.72, getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.72, getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.72, getAvgReliability(), ATTRIBUTE_DELTA);
		bos.reset();
	}
	
	@Test
	public void test_process_time_test_tcr_loop_geometric_small_1() throws EpcLoaderException {
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_tcr_loop_geometric_small_1.epml");
		assertNotNull(epc);
		processor.process(epc, AttributeType.TIME);
		assertEquals(10, getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(22.86, getAvgTime(), ATTRIBUTE_DELTA);
		bos.reset();
		processor.process(epc, AttributeType.COST);
		assertEquals(100, getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(228.6, getAvgCost(), ATTRIBUTE_DELTA);
		bos.reset();
		processor.process(epc, AttributeType.RELIABILITY);
		assertEquals(0.9, getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8036, getAvgReliability(), ATTRIBUTE_DELTA);
		bos.reset();
	}
	
//	@Test
//	public void test_process_time_test_tcr_loop_geometric_small_2() throws EpcLoaderException {
//		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_tcr_loop_geometric_small_2.epml");
//		helper.setNoMax(false);
//		assertNotNull(epc);
//		processor.process(epc, AttributeType.TIME);
//		assertEquals(50, getMinTime(), ATTRIBUTE_DELTA);
//		assertEquals(90, getAvgTime(), ATTRIBUTE_DELTA);
//		bos.reset();
//		processor.process(epc, AttributeType.COST);
//		assertEquals(300, getMinCost(), ATTRIBUTE_DELTA);
//		assertEquals(300, getMaxCost(), ATTRIBUTE_DELTA);
//		assertEquals(300, getAvgCost(), ATTRIBUTE_DELTA);
//		bos.reset();
//		processor.process(epc, AttributeType.RELIABILITY);
//		assertEquals(0.72, getMinReliability(), ATTRIBUTE_DELTA);
//		assertEquals(0.72, getMaxReliability(), ATTRIBUTE_DELTA);
//		assertEquals(0.72, getAvgReliability(), ATTRIBUTE_DELTA);
//		bos.reset();
//	}
	
	@Test
	public void test_process_time_test_tcr_loop_tuple_small_1() throws EpcLoaderException {
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_tcr_loop_tuple_small_1.epml");
		assertNotNull(epc);
		processor.process(epc, AttributeType.TIME);
		assertEquals(10, getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(100.0, getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(20.98, getAvgTime(), ATTRIBUTE_DELTA);
		bos.reset();
		processor.process(epc, AttributeType.COST);
		assertEquals(100, getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(1000, getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(209.8, getAvgCost(), ATTRIBUTE_DELTA);
		bos.reset();
		processor.process(epc, AttributeType.RELIABILITY);
		assertEquals(0.3359, getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.8127, getAvgReliability(), ATTRIBUTE_DELTA);
		bos.reset();
	}
	
//	@Test
//	public void test_process_time_test_tcr_loop_tuple_small_2() throws EpcLoaderException {
//		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_tcr_loop_tuple_small_2.epml");
//		helper.setNoMax(false);
//		assertNotNull(epc);
//		processor.process(epc, AttributeType.TIME);
//		assertEquals(50, getMinTime(), ATTRIBUTE_DELTA);
//		assertEquals(170.0, getMaxTime(), ATTRIBUTE_DELTA);
//		assertEquals(88.0, getAvgTime(), ATTRIBUTE_DELTA);
//		bos.reset();
//		processor.process(epc, AttributeType.COST);
//		assertEquals(300, getMinCost(), ATTRIBUTE_DELTA);
//		assertEquals(300, getMaxCost(), ATTRIBUTE_DELTA);
//		assertEquals(300, getAvgCost(), ATTRIBUTE_DELTA);
//		bos.reset();
//		processor.process(epc, AttributeType.RELIABILITY);
//		assertEquals(0.72, getMinReliability(), ATTRIBUTE_DELTA);
//		assertEquals(0.72, getMaxReliability(), ATTRIBUTE_DELTA);
//		assertEquals(0.72, getAvgReliability(), ATTRIBUTE_DELTA);
//		bos.reset();
//	}
	
	@Test
	public void test_process_time_test_tcr_xor_small_1() throws EpcLoaderException {
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_tcr_xor_small_1.epml");
		assertNotNull(epc);
		processor.process(epc, AttributeType.TIME);
		assertEquals(10, getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(20, getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(18, getAvgTime(), ATTRIBUTE_DELTA);
		bos.reset();
		processor.process(epc, AttributeType.COST);
		assertEquals(100, getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(200, getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(180, getAvgCost(), ATTRIBUTE_DELTA);
		bos.reset();
		processor.process(epc, AttributeType.RELIABILITY);
		assertEquals(0.8, getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.9, getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.82, getAvgReliability(), ATTRIBUTE_DELTA);
		bos.reset();
	}
	

	@Test
	public void test_process_time_test_tcr_xorand_medium_3() throws EpcLoaderException {
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_tcr_xorand_medium_3.epml");
		assertNotNull(epc);
		processor.process(epc, AttributeType.TIME);
		assertEquals(26, getMinTime(), ATTRIBUTE_DELTA);
		assertEquals(27, getMaxTime(), ATTRIBUTE_DELTA);
		assertEquals(26.6, getAvgTime(), ATTRIBUTE_DELTA);
		bos.reset();
		processor.process(epc, AttributeType.COST);
		assertEquals(3.6, getMinCost(), ATTRIBUTE_DELTA);
		assertEquals(3.7, getMaxCost(), ATTRIBUTE_DELTA);
		assertEquals(3.66, getAvgCost(), ATTRIBUTE_DELTA);
		bos.reset();
		processor.process(epc, AttributeType.RELIABILITY);
		assertEquals(0.45, getMinReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.81, getMaxReliability(), ATTRIBUTE_DELTA);
		assertEquals(0.594, getAvgReliability(), ATTRIBUTE_DELTA);
		bos.reset();
	}
}
