package de.fh_zwickau.bumb_l_b.qos_tool.verifier;

import static org.junit.Assert.*;

import java.util.ArrayList;
import org.jbpt.pm.ControlFlow;
import org.jbpt.pm.FlowNode;
import org.jbpt.pm.NonFlowNode;
import org.jbpt.pm.epc.IEpc;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.fh_zwickau.bumb_l_b.qos_tool.Message;
import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper.AttributeType;
import de.fh_zwickau.bumb_l_b.qos_tool.loader.EpcLoaderException;
import de.fh_zwickau.bumb_l_b.qos_tool.loader.EpmlEpcLoader;
import de.fh_zwickau.bumb_l_b.qos_tool.printer.AbstractPrinter;
import de.fh_zwickau.bumb_l_b.qos_tool.printer.AbstractPrinter.MsgType;

public class EpcVerifierTest {
	private static final TestPrinter printer = new TestPrinter();
	private static EpmlEpcLoader loader;
	private EpcVerifier verifier;
	private IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc;

	@BeforeClass
	public static void setUpBeforeClass() throws Exception {
		loader = new EpmlEpcLoader();
	}

	@Before
	public void setUp() throws Exception {
		printer.reset();
		verifier = new EpcVerifier(printer);
	}
	
	private void loadProcessor(String name) {
		load(name, "processor");
	}

	private void load(String name) {
		load(name, "verifier");
	}
	
	private void load(String name, String subdir) {
		try {
			epc = loader.loadFile("testContent/"+subdir+"/"+name+".epml");
		} catch (EpcLoaderException e) {
			fail("Load file failed");
		}
	}
	
	private static class TestPrinter extends AbstractPrinter {
		ArrayList<String> out = new ArrayList<String>();
		
		public TestPrinter() {}
		
		public void reset() {
			out.clear();
		}
		
		@Override
		protected void printMessage(MsgType type, String message, String elementId) {
			out.add(formatMessage(type, message, elementId));
		}
		
		public static String formatMessage(MsgType type, String message, String elementId) {
			return String.format("%s: %s", type, message);
		}
		
		public String getMessage(int index) {
			if(out.isEmpty()) { return ""; }
			return out.get(index);
		}
	};
	
	@Test
	public void test_contains_or() throws EpcLoaderException {
		load("contains_or");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.model_contains_or"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_all_attributes_processable() throws EpcLoaderException {
		load("all_attributes_processable");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertTrue(verifier.getProcessableAttributeTypes().contains(AttributeType.TIME));
		assertTrue(verifier.getProcessableAttributeTypes().contains(AttributeType.COST));
		assertTrue(verifier.getProcessableAttributeTypes().contains(AttributeType.RELIABILITY));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_time_on_all() throws EpcLoaderException {
		load("time_on_all");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertTrue(verifier.getProcessableAttributeTypes().contains(AttributeType.TIME));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_time_on_none() throws EpcLoaderException {
		load("time_on_none");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.TIME));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_time_on_some() throws EpcLoaderException {
		load("time_on_some");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.TIME));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, String.format(Message.get("verifier.incomplete_qos_attribute"), "time", "time"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_cost_on_all() throws EpcLoaderException {
		load("cost_on_all");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertTrue(verifier.getProcessableAttributeTypes().contains(AttributeType.COST));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_cost_on_none() throws EpcLoaderException {
		load("cost_on_none");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.COST));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_cost_on_some() throws EpcLoaderException {
		load("cost_on_some");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.COST));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, String.format(Message.get("verifier.incomplete_qos_attribute"), "cost", "cost"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_reliability_on_all() throws EpcLoaderException {
		load("reliability_on_all");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertTrue(verifier.getProcessableAttributeTypes().contains(AttributeType.RELIABILITY));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_reliability_on_none() throws EpcLoaderException {
		load("reliability_on_none");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.RELIABILITY));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_reliability_on_some() throws EpcLoaderException {
		load("reliability_on_some");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.RELIABILITY));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, String.format(Message.get("verifier.incomplete_qos_attribute"), "reliability", "reliability"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_noutofm_n_lt_1() throws EpcLoaderException {
		load("noutofm_n_lt_1");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.n_lower_than_1"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_noutofm_n_gt_m() throws EpcLoaderException {
		load("noutofm_n_gt_m");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.n_greater_than_m"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_noutofm_n_not_integer() throws EpcLoaderException {
		load("noutofm_n_not_integer");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.n_not_integer"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_noutofm_n_on_split() throws EpcLoaderException {
		load("noutofm_n_on_split");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.n_on_split"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_noutofm_n_ok() throws EpcLoaderException {
		load("noutofm_n_ok");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_time_lt_0() throws EpcLoaderException {
		load("time_lt_0");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.TIME));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.time_lower_than_0"), null), printer.getMessage(0));
	}

	@Test
	public void test_time_ge_0() throws EpcLoaderException {
		load("time_ge_0");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertTrue(verifier.getProcessableAttributeTypes().contains(AttributeType.TIME));
		assertEquals("", printer.getMessage(0));
	}

	@Test
	public void test_cost_lt_0() throws EpcLoaderException {
		load("cost_lt_0");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.COST));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.cost_lower_than_0"), null), printer.getMessage(0));
	}

	@Test
	public void test_cost_ge_0() throws EpcLoaderException {
		load("cost_ge_0");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertTrue(verifier.getProcessableAttributeTypes().contains(AttributeType.COST));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_reliability_ok() throws EpcLoaderException {
		load("reliability_ok");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertTrue(verifier.getProcessableAttributeTypes().contains(AttributeType.RELIABILITY));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_reliability_gt_1() throws EpcLoaderException {
		load("reliability_gt_1");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.RELIABILITY));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.reliability_limits_exceeded"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_reliability_le_0() throws EpcLoaderException {
		load("reliability_le_0");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.RELIABILITY));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.reliability_limits_exceeded"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_reliability_lt_0() throws EpcLoaderException {
		load("reliability_lt_0");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.RELIABILITY));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.reliability_limits_exceeded"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_ok() throws EpcLoaderException {
		load("probability_ok");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}

	@Test
	public void test_probability_gt_1() throws EpcLoaderException {
		load("probability_gt_1");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_limits_exceeded"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_ge_1() throws EpcLoaderException {
		load("probability_ge_1");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_limits_exceeded"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_le_0() throws EpcLoaderException {
		load("probability_le_0");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_limits_exceeded"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_lt_0() throws EpcLoaderException {
		load("probability_lt_0");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_limits_exceeded"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_list_ok() throws EpcLoaderException {
		load("probability_list_ok");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}

	@Test
	public void test_probability_list_gt_1() throws EpcLoaderException {
		load("probability_list_gt_1");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_list_limits_exceeded"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_list_ge_1() throws EpcLoaderException {
		load("probability_list_ge_1");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_list_limits_exceeded"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_list_le_0() throws EpcLoaderException {
		load("probability_list_le_0");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_list_limits_exceeded"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_list_lt_0() throws EpcLoaderException {
		load("probability_list_lt_0");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_list_limits_exceeded"), null), printer.getMessage(0));
	}

	@Test
	public void test_probability_list_not_descending() throws EpcLoaderException {
		load("probability_list_not_descending");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.WARNING, Message.get("verifier.probability_list_not_descending"), null), printer.getMessage(0));
	}

	@Test
	public void test_probability_mixed_single_list() throws EpcLoaderException {
		load("probability_mixed_single_list");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_single_and_list_mixed"), null), printer.getMessage(0));
	}

	@Test
	public void test_probability_list_num_edges_gt_2() throws EpcLoaderException {
		load("probability_list_num_edges_gt_2");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.xor_cycle_wrong_number_of_edges"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_list_num_edges_lt_2() throws EpcLoaderException {
		load("probability_list_num_edges_lt_2");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.xor_cycle_wrong_number_of_edges"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_list_multiple_used() throws EpcLoaderException {
		load("probability_list_multiple_used");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_list_more_than_one_list"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_incomplete2() throws EpcLoaderException {
		load("probability_incomplete2");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_value_missing"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_sum_gt_1() throws EpcLoaderException {
		load("probability_sum_gt_1");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_single_sum_fault"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_sum_lt_1() throws EpcLoaderException {
		load("probability_sum_lt_1");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_single_sum_fault"), null), printer.getMessage(0));
	}

//	moved to processor
//	@Test
//	public void test_probability_cycle_num_edges_gt_2() throws EpcLoaderException {
//		load("probability_cycle_num_edges_gt_2");
//		assertFalse("epc should be rejected", verifier.verify(epc));
//		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.xor_cycle_wrong_number_of_edges"), null), printer.getMessage(0));
//	}

//	moved to processor
//	@Test
//	public void test_probability_list_on_wrong_edge() throws EpcLoaderException {
//		load("probability_list_on_wrong_edge");
//		assertFalse("epc should be rejected", verifier.verify(epc));
//		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_list_on_wrong_edge"), null), printer.getMessage(0));
//	}

//	moved to processor
//	@Test
//	public void test_probability_num_cycles_gt_1() throws EpcLoaderException {
//		load("probability_num_cycles_gt_1");
//		assertFalse("epc should be rejected", verifier.verify(epc));
//		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.xor_more_than_one_cycle"), null), printer.getMessage(0));
//	}
	
//	moved to processor
//	@Test
//	public void test_probability_list_without_cycle() throws EpcLoaderException {
//		load("probability_list_without_cycle");
//		assertFalse("epc should be rejected", verifier.verify(epc));
//		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_list_on_wrong_edge"), null), printer.getMessage(0));
//	}
	
	@Test
	public void test_time_not_numeric() throws EpcLoaderException {
		load("time_not_numeric");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.TIME));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, String.format(Message.get("verifier.attribute_value_not_numeric"), "time"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_cost_not_numeric() throws EpcLoaderException {
		load("cost_not_numeric");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.COST));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, String.format(Message.get("verifier.attribute_value_not_numeric"), "cost"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_reliability_not_numeric() throws EpcLoaderException {
		load("reliability_not_numeric");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertFalse(verifier.getProcessableAttributeTypes().contains(AttributeType.RELIABILITY));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, String.format(Message.get("verifier.attribute_value_not_numeric"), "reliability"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_noutofm_n_not_numeric() throws EpcLoaderException {
		load("noutofm_n_not_numeric");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.n_not_integer"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_not_numeric() throws EpcLoaderException {
		load("probability_not_numeric");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_not_numeric_or_tuple"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_list_contains_non_numeric() throws EpcLoaderException {
		load("probability_list_contains_non_numeric");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_list_contains_not_numeric"), null), printer.getMessage(0));
	}
	
	@Test
	public void test_probability_list_is_empty() throws EpcLoaderException {
		load("probability_list_is_empty");
		assertFalse("epc should be rejected", verifier.verify(epc));
		assertEquals(TestPrinter.formatMessage(MsgType.ERROR, Message.get("verifier.probability_list_is_empty"), null), printer.getMessage(0));
	}
	
	//Processor testcases
	@Test
	public void test_minimal() throws EpcLoaderException {
		loadProcessor("minimal");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_linear1f() throws EpcLoaderException {
		loadProcessor("linear1f");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}

	@Test
	public void test_linear3f() throws EpcLoaderException {
		loadProcessor("linear3f");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_1path_1empty() throws EpcLoaderException {
		loadProcessor("and_1path_1empty");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_1path() throws EpcLoaderException {
		loadProcessor("and_1path");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_2path_1empty_1outof2() throws EpcLoaderException {
		loadProcessor("and_2path_1empty_1outof2");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_2path_1empty() throws EpcLoaderException {
		loadProcessor("and_2path_1empty");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_2path_1nestedand() throws EpcLoaderException {
		loadProcessor("and_2path_1nestedand");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_2path_1nestedlinear() throws EpcLoaderException {
		loadProcessor("and_2path_1nestedlinear");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_2path_2nestedand() throws EpcLoaderException {
		loadProcessor("and_2path_2nestedand");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_2path_2nestedandlinear() throws EpcLoaderException {
		loadProcessor("and_2path_2nestedandlinear");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_2path() throws EpcLoaderException {
		loadProcessor("and_2path");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_5path_1empty() throws EpcLoaderException {
		loadProcessor("and_5path_1empty");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_5path_1outof5() throws EpcLoaderException {
		loadProcessor("and_5path_1outof5");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_5path_2empty() throws EpcLoaderException {
		loadProcessor("and_5path_2empty");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_5path_3outof5() throws EpcLoaderException {
		loadProcessor("and_5path_3outof5");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_5path_5outof5() throws EpcLoaderException {
		loadProcessor("and_5path_5outof5");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_and_5path() throws EpcLoaderException {
		loadProcessor("and_5path");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_cycle_single_backempty() throws EpcLoaderException {
		loadProcessor("cycle_single_backempty");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_cycle_single_mainempty() throws EpcLoaderException {
		loadProcessor("cycle_single_mainempty");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_cycle_single() throws EpcLoaderException {
		loadProcessor("cycle_single");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test(timeout=10000)
	public void test_cycle_tuple1() throws EpcLoaderException {
		loadProcessor("cycle_tuple1");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test(timeout=10000)
	public void test_cycle_tuple10() throws EpcLoaderException {
		loadProcessor("cycle_tuple10");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test(timeout=10000)
	public void test_cycle_tuple3_backempty() throws EpcLoaderException {
		loadProcessor("cycle_tuple3_backempty");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_cycle_tuple3_mainempty() throws EpcLoaderException {
		loadProcessor("cycle_tuple3_mainempty");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_cycle_tuple3() throws EpcLoaderException {
		loadProcessor("cycle_tuple3");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_xor_2path_1empty() throws EpcLoaderException {
		loadProcessor("xor_2path_1empty");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_xor_2path_1nestedand() throws EpcLoaderException {
		loadProcessor("xor_2path_1nestedand");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_xor_2path_1nestedand2() throws EpcLoaderException {
		loadProcessor("xor_2path_1nestedand2");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_xor_2path_1nestedand3() throws EpcLoaderException {
		loadProcessor("xor_2path_1nestedand3");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_xor_2path_1nestedcycle() throws EpcLoaderException {
		loadProcessor("xor_2path_1nestedcycle");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_xor_2path_1nestedcycle2() throws EpcLoaderException {
		loadProcessor("xor_2path_1nestedcycle2");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_xor_2path_1nestedcycle3() throws EpcLoaderException {
		loadProcessor("xor_2path_1nestedcycle3");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_xor_2path_1nestedlinear() throws EpcLoaderException {
		loadProcessor("xor_2path_1nestedlinear");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_xor_2path_1nestedxor() throws EpcLoaderException {
		loadProcessor("xor_2path_1nestedxor");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_xor_2path_1nestedxor2() throws EpcLoaderException {
		loadProcessor("xor_2path_1nestedxor2");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_xor_2path_1nestedxor3() throws EpcLoaderException {
		loadProcessor("xor_2path_1nestedxor3");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}
	
	@Test
	public void test_xor_2path_3nestedcycle_weird() throws EpcLoaderException {
		loadProcessor("xor_2path_3nestedcycle_weird");
		assertTrue("epc should be accepted", verifier.verify(epc));
		assertEquals("", printer.getMessage(0));
	}

	
}

