package de.fh_zwickau.bumb_l_b.qos_tool.verifier;

import static org.junit.Assert.*;

import java.util.Set;

import org.jbpt.pm.ControlFlow;
import org.jbpt.pm.FlowNode;
import org.jbpt.pm.NonFlowNode;
import org.jbpt.pm.epc.IEpc;
import org.junit.Before;
import org.junit.Test;

import de.fh_zwickau.bumb_l_b.qos_tool.customJBPT.AttributeHelper.AttributeType;
import de.fh_zwickau.bumb_l_b.qos_tool.loader.EpcLoaderException;
import de.fh_zwickau.bumb_l_b.qos_tool.loader.EpmlEpcLoader;
import de.fh_zwickau.bumb_l_b.qos_tool.printer.StaticPrinterFactory;
import de.fh_zwickau.bumb_l_b.qos_tool.printer.StaticPrinterFactory.PrinterType;

public class TestEpcVerifier {

	private EpcVerifier verifier;
	private EpmlEpcLoader loader;
	
	@Before
	public void setUp() throws Exception {
		loader = new EpmlEpcLoader();
		verifier = new EpcVerifier(StaticPrinterFactory.getPrinterInstance(PrinterType.DEFAULT));		
	}

	@Test
	public void testEpcCheck() throws EpcLoaderException 
	{	
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/testkette.epml");
		assertFalse(verifier.verify(epc));		
		
	}

	@Test
	public void testVerify_tcr_linear_small_1() throws EpcLoaderException {	
		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_tcr_linear_small_1.epml");
		assertTrue(verifier.verify(epc));
		Set<AttributeType> processable = verifier.getProcessableAttributeTypes();
		assertEquals(3, processable.size());
		assertTrue(processable.contains(AttributeType.COST));
		assertTrue(processable.contains(AttributeType.TIME));
		assertTrue(processable.contains(AttributeType.RELIABILITY));
		System.out.println("--------------------------------------------------------------------------------");
	}
	
	@Test
	public void testVerify_incomplete_attributes() throws EpcLoaderException {

		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_incomplete_attributes.epml");
		assertFalse(verifier.verify(epc));
		Set<AttributeType> processable = verifier.getProcessableAttributeTypes();
		assertEquals(0, processable.size());
		System.out.println("--------------------------------------------------------------------------------");
	}
	
	@Test
	public void testVerify_complete_att_cost_oob() throws EpcLoaderException {

		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_complete_att_cost_OOB.epml");
		assertFalse(verifier.verify(epc));
		Set<AttributeType> processable = verifier.getProcessableAttributeTypes();
		assertEquals(2, processable.size());
		assertTrue(processable.contains(AttributeType.TIME));
		assertTrue(processable.contains(AttributeType.RELIABILITY));		
		System.out.println("--------------------------------------------------------------------------------");
	}
	
	@Test
	public void testVerify_complete_att_time_oob() throws EpcLoaderException {

		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_complete_att_time_OOB.epml");
		assertFalse(verifier.verify(epc));
		Set<AttributeType> processable = verifier.getProcessableAttributeTypes();
		assertEquals(2, processable.size());
		assertTrue(processable.contains(AttributeType.COST));
		assertTrue(processable.contains(AttributeType.RELIABILITY));		
		System.out.println("--------------------------------------------------------------------------------");
	}
	
	@Test
	public void testVerify_complete_att_relia_oob() throws EpcLoaderException {

		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_complete_att_relia_OOB.epml");
		assertFalse(verifier.verify(epc));
		Set<AttributeType> processable = verifier.getProcessableAttributeTypes();
		assertEquals(2, processable.size());
		assertTrue(processable.contains(AttributeType.TIME));
		assertTrue(processable.contains(AttributeType.COST));
		System.out.println("--------------------------------------------------------------------------------");
	}
	
	@Test
	public void testVerify_complete_att_no_cost() throws EpcLoaderException {

		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_complete_att_no_cost.epml");
		assertFalse(verifier.verify(epc));
		Set<AttributeType> processable = verifier.getProcessableAttributeTypes();
		assertEquals(2, processable.size());
		assertTrue(processable.contains(AttributeType.TIME));
		assertTrue(processable.contains(AttributeType.RELIABILITY));	
		System.out.println("--------------------------------------------------------------------------------");
	}
	
	@Test
	public void testVerify_valid_xor() throws EpcLoaderException {

		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_valid_xor.epml");
		assertTrue(verifier.verify(epc));
		Set<AttributeType> processable = verifier.getProcessableAttributeTypes();
		assertEquals(1, processable.size());
		assertTrue(processable.contains(AttributeType.PROBABILITY));	
		System.out.println("--------------------------------------------------------------------------------");
	}
	
	@Test
	public void testVerify_xor_missing_attribute() throws EpcLoaderException {

		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_xor_missing_attribute.epml");
		assertFalse(verifier.verify(epc));
		Set<AttributeType> processable = verifier.getProcessableAttributeTypes();
		assertEquals(0, processable.size());
		assertFalse(processable.contains(AttributeType.PROBABILITY));	
		System.out.println("--------------------------------------------------------------------------------");
	}
	
	@Test
	public void testVerify_xor_over_1() throws EpcLoaderException {

		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_xor_over_1.epml");
		assertFalse(verifier.verify(epc));
		Set<AttributeType> processable = verifier.getProcessableAttributeTypes();
		assertEquals(0, processable.size());
		assertFalse(processable.contains(AttributeType.PROBABILITY));	
		System.out.println("--------------------------------------------------------------------------------");
	}
	
	@Test
	public void test_tcr_xorand_medium_1() throws EpcLoaderException {

		IEpc<ControlFlow<FlowNode>, FlowNode, NonFlowNode> epc = loader.loadFile("testContent/test_tcr_xorand_medium_1.epml");
		assertTrue(verifier.verify(epc));
		System.out.println("--------------------------------------------------------------------------------");
	}
}
